# $Id: qa_helpers.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import types
import re
import sys

class regexp(object):
    r'''
        >>> m = regexp(ur'(hi)\s*there', u'the msg', u'the prompt')
        >>> m
        <regexp u'the msg'[the prompt]/(hi)\s*there/>
        >>> m.msg
        u'the msg'
        >>> m.prompt
        u'the prompt'
        >>> m.match(u'hithere')
        u'hi'
    '''
    def __init__(self, regexp, msg=None, prompt=None):
        self.re = re.compile(regexp, re.UNICODE | re.VERBOSE)
        self.pattern = regexp
        self.msg = msg
        self.prompt = prompt

    def __repr__(self):
        if self.msg:
            if self.prompt:
                return '<regexp %r[%s]/%s/>' % \
                       (self.msg, str(self.prompt), self.pattern)
            return '<regexp %r/%s/>' % (self.msg, self.pattern)
        if self.prompt:
            return '<regexp [%s]/%s/>' % (self.prompt, self.pattern)
        return '<regexp /%s/>' % self.pattern

    def __getstate__(self):
        return self.pattern, self.msg, self.prompt

    def __setstate__(self, args):
        self.pattern, self.msg, self.prompt = args
        self.re = re.compile(self.pattern, re.UNICODE | re.VERBOSE)

    def match(self, str):
        m = self.re.match(str)
        if m and m.end() == len(str):
            if m.lastindex > 1: return m.groups()
            if m.lastindex == 1: return m.group(1)
            return str

class qmap(object):
    r'''
        >>> m = qmap(u'y', True)
        >>> m
        <qmap True = u'y'>
        >>> m.test
        u'y'
        >>> m.value
        True
    '''
    def __init__(self, test, value):
        self.test = test
        self.value = value

    def __repr__(self):
        return "<qmap %s = %s>" % (repr(self.value), repr(self.test))

if sys.version_info[0] < 3:
    def urepr(x):
        r'''
            >>> urepr(44)
            u'44'
            >>> tuple(urepr('hi\n'))
            (u"'", u'h', u'i', u'\\', u'n', u"'")
        '''
        if isinstance(x, types.StringTypes):
            return repr(x.encode('utf-8')).decode('utf-8')
        return unicode(x)
else:
    urepr = repr

def to_int(str):
    r'''
        >>> to_int(u' -34')
        -34
        >>> to_int(u' +43')
        43
        >>> to_int(u'43x')
        Traceback (most recent call last):
            ...
        ValueError: illegal integer: '43x'
    '''
    try:
        return int(str)
    except ValueError:
        raise ValueError(u"illegal integer: %s" % urepr(str))

def to_float(str):
    r'''
        >>> str(to_float(u' -34.444'))
        '-34.444'
        >>> str(to_float(u' +43'))
        '43.0'
        >>> to_float(u'43.3.3')
        Traceback (most recent call last):
            ...
        ValueError: illegal floating point number: '43.3.3'
    '''
    try:
        return float(str)
    except ValueError:
        raise ValueError(u"illegal floating point number: %s" % urepr(str))

def to_number(str):
    r'''
        >>> str(to_number(u' -34.444'))
        '-34.444'
        >>> to_number(u' +43')
        43
        >>> to_number(u'43.3.3')
        Traceback (most recent call last):
            ...
        ValueError: illegal number: '43.3.3'
    '''
    try:
        return to_int(str)
    except ValueError:
        try:
            return to_float(str)
        except ValueError:
            raise ValueError(u"illegal number: %s" % urepr(str))

def to_tuple(str, conv_fn=None, test=None, separator=','):
    r'''
        >>> to_tuple(u'1, 2,3', to_int)
        (1, 2, 3)
        >>> to_tuple(u'1, 2.5, -7e3', to_number)
        (1, 2.5, -7000.0)
        >>> to_tuple(u'43', to_number)
        (43,)
        >>> to_tuple(u'1,43.3.3', to_number)
        Traceback (most recent call last):
            ...
        ValueError: illegal number: '43.3.3'
    '''

    def conv_element(elem):
        elem = elem.strip()
        if conv_fn: elem = conv_fn(elem)
        if test: elem = match(elem, test)
        return elem

    return tuple(conv_element(elem) for elem in str.split(separator))

def msg_for(test, type):
    r'''
        >>> msg_for(None, int)
        >>> msg_for(regexp(u'', u'the msg'), int)
        u'the msg'
        >>> msg_for(qmap(44, True), int)
        u'44'
        >>> msg_for(slice(3, 55), int)
        u'between 3 and 55'
        >>> msg_for(slice(None, 55), int)
        u'<= 55'
        >>> msg_for(slice(3, None), int)
        u'>= 3'
        >>> msg_for(slice(None, None), int)
        u''
        >>> msg_for(slice(3, 55), str)
        u'between 3 and 55 characters'
        >>> msg_for(slice(None, 55), str)
        u'<= 55 characters'
        >>> msg_for(slice(3, None), str)
        u'>= 3 characters'
        >>> msg_for(slice(None, None), str)
        u''
        >>> msg_for((slice(3, 5), True), str)
        u'between 3 and 5 characters or True'
        >>> msg_for(True, str)
        u'True'
    '''
    if test is None: return None
    if isinstance(test, regexp): return test.msg
    if isinstance(test, qmap): return msg_for(test.test, type)
    if isinstance(test, slice):
        if test.start is None:
            if test.stop is not None: ans = u"<= %d" % test.stop
            else: ans = u""
        elif test.stop is None:
            ans = u">= %d" % test.start
        else:
            ans = u"between %d and %d" % (test.start, test.stop)
        if (type == str or type == unicode) and ans: ans += u' characters'
        return ans
    if isinstance(test, (tuple, list)):
        return u' or '.join(filter(None, (msg_for(test_i, type)
                                          for test_i in test)))
    return urepr(test)

def match_prompt(test, type, format, default=u''):
    r'''
        >>> match_prompt(None, int, u' [%s] ')
        u''
        >>> match_prompt(regexp(u'', u'', u'the prompt'), int, u' [%s] ')
        u' [the prompt] '
        >>> match_prompt(qmap(44, True), int, u' [%s] ')
        u' [44] '
        >>> match_prompt(slice(3, 55), int, u' [%s] ')
        u' [3-55] '
        >>> match_prompt(slice(None, 55), int, u' [%s] ')
        u' [max 55] '
        >>> match_prompt(slice(3, None), int, u' [%s] ')
        u' [min 3] '
        >>> match_prompt(slice(None, None), int, u' [%s] ', u'foo')
        u'foo'
        >>> match_prompt(slice(3, 55), str, u' [%s] ')
        u' [len: 3-55] '
        >>> match_prompt(slice(None, 55), str, u' [%s] ')
        u' [len <= 55] '
        >>> match_prompt(slice(3, None), str, u' [%s] ')
        u' [len >= 3] '
        >>> match_prompt(slice(None, None), str, u' [%s] ')
        u''
        >>> match_prompt((slice(3, 5), True), str, u' [%s] ')
        u' [len: 3-5 or True] '
        >>> match_prompt(True, str, u' [%s] ')
        u' [True] '
    '''

    def prompt_body(test, type):
        if test is None: return None
        if isinstance(test, regexp): return test.prompt
        if isinstance(test, qmap): return prompt_body(test.test, type)
        if isinstance(test, slice):
            if test.start is None:
                if test.stop is not None:
                    if issubclass(type, types.StringTypes):
                        return u"len <= %d" % test.stop
                    else:
                        return u"max %d" % test.stop
                else: return u""
            elif test.stop is None:
                if issubclass(type, types.StringTypes):
                    return u"len >= %d" % test.start
                else:
                    return u"min %d" % test.start
            else:
                if issubclass(type, types.StringTypes):
                    return u"len: %d-%d" % (test.start, test.stop)
                else:
                    return u"%d-%d" % (test.start, test.stop)
        if isinstance(test, (tuple, list)):
            return u' or '.join(filter(None, (prompt_body(test_i, type)
                                              for test_i in test)))
        return urepr(test)

    body = prompt_body(test, type)
    if body: return format % body
    return default

def match(ans, test):
    r'''
        >>> match(u'foobar', None)
        u'foobar'
        >>> match(u'hithere', regexp(ur'(hi)\s*there', u'hi there'))
        u'hi'
        >>> match(u'hi mom', regexp(ur'(hi)\s*there', u'hi there'))
        Traceback (most recent call last):
            ...
        ValueError: hi there
        >>> match(u'y', qmap(u'y', True))
        True
        >>> match(2, qmap(slice(3, 5), True))
        Traceback (most recent call last):
            ...
        ValueError: between 3 and 5
        >>> match(3, slice(3,5))
        3
        >>> match(2, slice(3,5))
        Traceback (most recent call last):
            ...
        ValueError: between 3 and 5
        >>> match(2, (slice(3,5), slice(5,10), 2))
        2
        >>> match(2, 2)
        2
    '''
    if test is None: return ans
    if isinstance(test, regexp):
        ans = test.match(ans)
        if ans is not None: return ans
    if isinstance(test, qmap):
        match(ans, test.test)   # raises ValueError if it doesn't match
        return test.value
    elif isinstance(test, slice):
        if isinstance(ans, types.StringTypes): value = len(ans)
        else: value = ans
        if (test.start is None or value >= test.start) and \
           (test.stop is None or value <= test.stop):
            return ans
    elif isinstance(test, (tuple, list)):
        for test_i in test:
            try:
                return match(ans, test_i)
            except ValueError:
                pass
    elif test == ans: return ans
    raise ValueError(msg_for(test, type(ans)))




# $Id: rule_base.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import itertools
from pyke import knowledge_base

class StopProof(Exception): pass

class stopIteratorContext(object):
    def __init__(self, rule_base, iterator_context):
        self.rule_base = rule_base
        self.context = iterator_context

    def __enter__(self):
        return stopIterator(self.rule_base, self.context.__enter__())

    def __exit__(self, type, value, tb):
        self.context.__exit__(type, value, tb)

class stopIterator(object):
    def __init__(self, rule_base, iterator):
        self.rule_base = rule_base
        self.iterator = iter(iterator)

    def __iter__(self): return self

    def next(self):
        if self.iterator:
            try:
                return self.iterator.next()
            except StopProof:
                self.iterator = None
                self.rule_base.num_bc_rule_failures += 1
        raise StopIteration

class chain_context(object):
    def __init__(self, outer_it):
        self.outer_it = outer_iterable(outer_it)

    def __enter__(self):
        return itertools.chain.from_iterable(self.outer_it)

    def __exit__(self, type, value, tb): self.outer_it.close()

class outer_iterable(object):
    def __init__(self, outer_it):
        self.outer_it = iter(outer_it)
        self.inner_it = None

    def __iter__(self): return self

    def close(self):
        if hasattr(self.inner_it, '__exit__'):
            self.inner_it.__exit__(None, None, None)
        elif hasattr(self.inner_it, 'close'): self.inner_it.close()
        if hasattr(self.outer_it, 'close'): self.outer_it.close()

    def next(self):
        ans = self.outer_it.next()
        if hasattr(ans, '__enter__'):
            self.inner_it = ans
            return ans.__enter__()
        ans = iter(ans)
        self.inner_it = ans
        return ans

class rule_base(knowledge_base.knowledge_base):
    def __init__(self, engine, name, parent = None, exclude_list = ()):
        super(rule_base, self).__init__(engine, name, rule_list, False)
        if name in engine.rule_bases:
            raise AssertionError("rule_base %s already exists" % name)
        if name in engine.knowledge_bases:
            raise AssertionError("name clash between rule_base '%s' and "
                                 "fact_base '%s'" % (name, name))
        engine.rule_bases[name] = self
        self.fc_rules = []
        self.parent = parent
        self.exclude_set = frozenset(exclude_list)
        self.rules = {}         # {name: rule}

    def add_fc_rule(self, fc_rule):
        if fc_rule.name in self.rules:
            raise AssertionError("%s rule_base: duplicate rule name: %s" %
                                 (self.name, fc_rule.name))
        self.rules[fc_rule.name] = fc_rule
        self.fc_rules.append(fc_rule)

    def add_bc_rule(self, bc_rule):
        if bc_rule.name in self.rules:
            raise AssertionError("%s rule_base: duplicate rule name: %s" %
                                 (self.name, bc_rule.name))
        self.rules[bc_rule.name] = bc_rule
        self.get_entity_list(bc_rule.goal_name).add_bc_rule(bc_rule)

    def init2(self):
        if not self.initialized:
            self.initialized = True
            if self.parent:
                parent = self.engine.rule_bases.get(self.parent)
                if parent is None:
                    raise KeyError("rule_base %s: parent %s not found" % \
                                   (self.name, self.parent))
                self.parent = parent
                self.parent.init2()
                self.root_name = self.parent.root_name
            else:
                self.root_name = self.name
            self.reset()

    def derived_from(self, rb):
        parent = self.parent
        while parent:
            if parent == rb: return True
            parent = parent.parent
        return False

    def register_fc_rules(self, stop_at_rb):
        rb = self
        while rb is not stop_at_rb:
            for fc_rule in rb.fc_rules: fc_rule.register_rule()
            if not rb.parent: break
            rb = rb.parent

    def run_fc_rules(self, stop_at_rb):
        rb = self
        while rb is not stop_at_rb:
            for fc_rule in rb.fc_rules: fc_rule.run()
            if not rb.parent: break
            rb = rb.parent

    def activate(self):
        current_rb = self.engine.knowledge_bases.get(self.root_name)
        if current_rb:
            assert self.derived_from(current_rb), \
                   "%s.activate(): not derived from current rule_base, %s" % \
                   (self.name, current_rb.name)
        self.engine.knowledge_bases[self.root_name] = self
        self.register_fc_rules(current_rb)
        self.run_fc_rules(current_rb)

    def reset(self):
        if self.root_name in self.engine.knowledge_bases:
            del self.engine.knowledge_bases[self.root_name]
        for fc_rule in self.fc_rules: fc_rule.reset()
        self.num_fc_rules_triggered = 0
        self.num_fc_rules_rerun = 0
        self.num_prove_calls = 0
        self.num_bc_rules_matched = 0
        self.num_bc_rule_successes = 0
        self.num_bc_rule_failures = 0

    def gen_rule_lists_for(self, goal_name):
        rule_base = self
        while True:
            rl = rule_base.entity_lists.get(goal_name)
            if rl: yield rl
            if rule_base.parent and goal_name not in rule_base.exclude_set:
                rule_base = rule_base.parent
            else:
                break

    def prove(self, bindings, pat_context, goal_name, patterns):
        self.num_prove_calls += 1
        return stopIteratorContext(self,
                   chain_context(
                       rl.prove(bindings, pat_context, patterns)
                       for rl in self.gen_rule_lists_for(goal_name)))

    def print_stats(self, f):
        f.write("%s: %d fc_rules, %d triggered, %d rerun\n" %
                (self.name, len(self.fc_rules), self.num_fc_rules_triggered,
                 self.num_fc_rules_rerun))
        num_bc_rules = sum(rule_list.num_bc_rules()
                             for rule_list in self.entity_lists.itervalues())
        f.write("%s: %d bc_rules, %d goals, %d rules matched\n" %
                (self.name, num_bc_rules, self.num_prove_calls,
                 self.num_bc_rules_matched))
        f.write("%s  %d successes, %d failures\n" %
                (' ' * len(self.name), self.num_bc_rule_successes,
                 self.num_bc_rule_failures))
        if self.parent: self.parent.print_stats(f)

    def trace(self, rule_name):
        for rule_list in self.entity_lists.itervalues():
            if rule_list.trace(rule_name): return
        raise KeyError("trace: rule %s not found" % rule_name)

    def untrace(self, rule_name):
        for rule_list in self.entity_lists.itervalues():
            if rule_list.untrace(rule_name): return
        raise KeyError("untrace: rule %s not found" % rule_name)

class rule_list(knowledge_base.knowledge_entity_list):
    def __init__(self, name):
        self.name = name
        self.bc_rules = []

    def add_bc_rule(self, bc_rule):
        self.bc_rules.append(bc_rule)

    def prove(self, bindings, pat_context, patterns):
        """ Returns a context manager for a generator that binds patterns to
            successively proven goals, yielding the plan (or None, if no plan)
            for each successful match.  Undoes bindings upon continuation,
            so that no bindings remain at StopIteration.
        """
        return chain_context(
                   bc_rule.bc_fn(bc_rule, patterns, pat_context)
                   for bc_rule in self.bc_rules)

    def num_bc_rules(self):
        return len(self.bc_rules)

    def trace(self, rule_name):
        for bc_rule in self.bc_rules:
            if bc_rule.name == rule_name:
                bc_rule.trace()
                return True
        return False

    def untrace(self, rule_name):
        for bc_rule in self.bc_rules:
            if bc_rule.name == rule_name:
                bc_rule.untrace()
                return True
        return False




# $Id: kqb_parser.py 49507964ae64 2010-03-27 mtnyogi $
# coding=utf-8
# 
# Copyright © 2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import with_statement
from string import Template
import re
import os.path
from pyke import question_base
from pyke import user_question
from pyke import qa_helpers
from pyke.krb_compiler import scanner

class kqb_parser(object):
    blank_line = re.compile(ur'(\s*#)|(\s*$)', re.UNICODE)
    tokenizer = re.compile(ur''' [ \t\f\r\v]* (?: \#.* )? (?:
            (["']) (?P<str> (?: \\. | .)*? ) \1 |        # this must be first!
            [[] (?P<prompt> (?: \\. | .)*? ) []] |
            [$] (?P<param> [a-zA-Z_] [a-zA-Z_0-9]* ) |
            / (?P<regexp1> (?: \\. | .)*? ) / |
            /// (?P<regexp2> (?: \\. | \n | .)*? ) /// | # FIX: this won't work!
            (?P<const> True | False | None ) |
            (?P<id> [a-zA-Z_] [a-zA-Z_0-9]* ) |
            (?P<number> (?: \d+ (?: \.\d* )? |
                            \.\d+ )
                        (?: [eE][-+]?\d+ )? ) |
            (?P<lparen> [(] ) |
            (?P<rparen> [)] ) |
            (?P<comma> , ) |
            (?P<bar> [|] ) |
            (?P<bang> ! ) |
            (?P<equal> = ) |
            (?P<hyphen> - ) |
            (?P<colon> : ) 
        ) [ \t\f\r\v]* (?: \#.* )? ''', re.UNICODE | re.VERBOSE)
    pushed_token = None

    def __init__(self, f):
        # f needs readline() and name.
        self.f = f
        self.lineno = 0
        self.line = ''
        self.column = 0
        self.eof = False

    def readline(self):
        r'''
            >>> from StringIO import StringIO
            >>> p = kqb_parser(StringIO("""
            ... line 1
            ...     # this should be ignored
            ...  line 2
            ...
            ...   line 3
            ... """))
            >>> p.readline()
            >>> p.indent, p.line, p.lineno, p.column
            (0, 'line 1', 2, 0)
            >>> p.readline()
            >>> p.indent, p.line, p.lineno, p.column
            (1, ' line 2', 4, 1)
            >>> p.readline()
            >>> p.indent, p.line, p.lineno, p.column
            (2, '  line 3', 6, 2)
            >>> p.eof
            False
            >>> p.readline()
            >>> p.eof
            True
        '''
        while True:
            line = self.f.readline()
            if line == '':
                self.eof = True
                break
            self.lineno += 1
            line = line.rstrip('\n')
            if not self.blank_line.match(line):
                self.indent, self.column = scanner.count_indent(line)
                self.line = line
                break

    def SyntaxError(self, msg, last_token=True):
        if last_token:
            raise SyntaxError(msg,
                              (self.f.name, self.last_lineno,
                               self.last_column + 1, self.last_line))
        raise SyntaxError(msg,
                          (self.f.name, self.lineno, self.column + 1,
                           self.line))

    def push_token(self):
        #print "push_token:", self.last_token  # FIX
        self.pushed_token = self.last_token
        self.pushed_indent = self.indent
        self.pushed_column = self.column
        self.indent = self.last_indent
        self.column = self.last_column

    def get_token(self, check_token=None):
        r'''
            >>> from StringIO import StringIO
            >>> f = StringIO(r"""
            ...   line 1=2.5: ( /\s* /, $foo) # comment
            ... ,|!-True  "hi\n"  [mom]
            ... """)
            >>> f.name = 'StringIO'
            >>> p = kqb_parser(f)
            >>> p.get_token()
            ('id', 'line')
            >>> p.get_token()
            ('number', 1)
            >>> p.get_token()
            ('equal', None)
            >>> p.get_token()
            ('number', 2.5)
            >>> p.get_token()
            ('colon', None)
            >>> p.get_token()
            ('lparen', None)
            >>> p.get_token()
            ('regexp', '\\s* ')
            >>> p.get_token()
            ('comma', None)
            >>> p.get_token()
            ('param', 'foo')
            >>> p.get_token()
            ('rparen', None)
            >>> p.get_token()
            ('comma', None)
            >>> p.get_token()
            ('bar', None)
            >>> p.get_token()
            ('bang', None)
            >>> p.get_token()
            ('hyphen', None)
            >>> p.get_token()
            ('const', True)
            >>> p.get_token()
            ('str', 'hi\n')
            >>> p.get_token()
            ('prompt', 'mom')
            >>> p.get_token()
            (None, None)
        '''
        if self.pushed_token:
            ans = self.pushed_token
            self.indent = self.pushed_indent
            self.column = self.pushed_column
            self.pushed_token = self.pushed_column = self.pushed_indent = None
            if check_token and check_token != ans[0]:
                self.SyntaxError("expected %s, got %s" % (check_token, ans[0]))
            #print "get_token: returning pushed_token", ans  # FIX
            return ans
        if self.column < len(self.line): self.skip_spaces()
        if self.column >= len(self.line):
            self.readline()
            if self.eof:
                self.last_token = None, None
                #print "get_token: returning EOF"  # FIX
                return self.last_token
        self.last_line = self.line
        self.last_lineno = self.lineno
        self.last_column = self.column
        self.last_indent = self.indent
        match = self.tokenizer.match(self.line, self.column)
        if not match: self.SyntaxError("Scanner error: no legal token")
        token = match.lastgroup
        chars = match.group(token)
        end = match.end()
        indent, ignore = scanner.count_indent(self.line[self.column:end], True)
        self.indent += indent
        self.column = end
        if token == 'str' or token == 'prompt':
            value = scanner.unescape(chars)
        elif token == 'const':
            value = eval(chars)
        elif token == 'number':
            try:
                value = int(chars)
            except ValueError:
                value = float(chars)
        elif token == 'param' or token == 'id':
            value = chars
        elif token == 'regexp1' or token == 'regexp2':
            # FIX
            token = 'regexp'
            value = chars
            self.lineno += chars.count('\n')
            last_nl = chars.rfind('\n')
            if last_nl >= 0:
                self.column = len(chars) - last_nl + 4
        else:
            value = None
        if check_token and check_token != token:
            self.SyntaxError("expected %s, got %s" % (check_token, token))
        self.last_token = str(token), value
        #print "get_token: returning", self.last_token  # FIX
        return self.last_token

    def get_block_string(self, stop=None, hanging=False, ending_newlines=False):
        r'''
            >>> from StringIO import StringIO
            >>> f = StringIO(r"""
            ...     line 1 # comment
            ...        more stuff
            ...     last line
            ... blah    hanging line 1
            ...
            ...         line 2
            ...           indented
            ...         last line
            ...         ! the end !
            ... """)
            >>> f.name = 'StringIO'
            >>> p = kqb_parser(f)
            >>> p.get_block_string()
            u'line 1 # comment\n   more stuff\nlast line'
            >>> p.column = 4
            >>> p.indent = 4
            >>> p.get_block_string('!', True)
            u'hanging line 1\n\nline 2\n  indented\nlast line'
            >>> f = StringIO(r"""
            ...     ! line 1 # comment
            ...          more stuff
            ...       last line
            ... blah
            ... """)
            >>> f.name = 'StringIO'
            >>> p = kqb_parser(f)
            >>> p.readline()
            >>> p.get_token('bang')
            ('bang', None)
            >>> p.get_block_string(hanging=True)
            u'line 1 # comment\n   more stuff\nlast line'
        '''
        if hanging:
            indent, more_chars = \
                scanner.count_indent(self.line[self.column:])
            self.indent += indent
            self.column += more_chars
        else:
            self.readline()
        indent = self.indent
        if self.eof: self.SyntaxError("expected block string, got EOF", False)
        rest_line = self.line[self.column:]
        if self.blank_line.match(rest_line):
            ans = []
        else:
            ans = [rest_line]
        while not self.eof:
            last_lineno = self.lineno
            self.readline()
            if ending_newlines:
                for i in range(self.lineno - last_lineno - 1): ans.append('')
            if self.eof or self.indent < indent or \
               stop and self.line[self.column:].startswith(stop):
                break
            if not ending_newlines:
                for i in range(self.lineno - last_lineno - 1): ans.append('')
            ans.append(' ' * (self.indent - indent) + self.line[self.column:])
        if not ans: self.SyntaxError("expected block string", False)
        return u'\n'.join(scanner.unescape(str) for str in ans)

    def parse_simple_match(self):
        token, value = self.get_token()
        if token == 'str' or token == 'id' or token == 'number' or \
           token == 'const':
            next_token, next_value = self.get_token()
            if next_token == 'prompt' and token == 'str':
                final_token, final_value = self.get_token('regexp')
                return qa_helpers.regexp(final_value, value, next_value)
            if next_token == 'regexp' and token == 'str':
                return qa_helpers.regexp(next_value, value)
            if next_token == 'equal':
                return qa_helpers.qmap(self.parse_simple_match(), value)
            if next_token == 'hyphen' and token == 'number':
                final_token, final_value = self.get_token()
                if final_token == 'number':
                    return slice(value, final_value)
                self.push_token()
                return slice(value, None)
            self.push_token()
            return value
        if token == 'prompt':
            next_token, next_value = self.get_token('regexp')
            return qa_helpers.regexp(next_value, prompt=value)
        if token == 'lparen':
            ans = self.parse_match()
            self.get_token('rparen')
            return ans
        if token == 'regexp':
            return qa_helpers.regexp(value)
        if token == 'hyphen':
            next_token, next_value = self.get_token('number')
            return slice(None, next_value)
        self.SyntaxError("expected match, got %s" % token)

    def parse_match(self):
        r'''
            >>> from StringIO import StringIO
            >>> def do(str):
            ...    ans = StringIO(str)
            ...    ans.name = 'StringIO'
            ...    return kqb_parser(ans).parse_match()
            >>> do(r'/reg\exp/ bob')
            <regexp /reg\exp/>
            >>> do(r'"msg"/reg\exp/ bob')
            <regexp 'msg'/reg\exp/>
            >>> do(r'[prompt]/reg\exp/ bob')
            <regexp [prompt]/reg\exp/>
            >>> do(r'"msg"[prompt]/reg\exp/ bob')
            <regexp 'msg'[prompt]/reg\exp/>
            >>> do(r"44 = id")
            <qmap 44 = 'id'>
            >>> do(r"-5 bob")
            slice(None, 5, None)
            >>> do(r"0-5 bob")
            slice(0, 5, None)
            >>> do(r"0- bob")
            slice(0, None, None)
            >>> do(r"/regexp/|44|3-5 bob")
            (<regexp /regexp/>, 44, slice(3, 5, None))
            >>> do(r"44 id")
            44
        '''
        ans = [self.parse_simple_match()]
        token, value = self.get_token()
        while token == 'bar':
            ans.append(self.parse_simple_match())
            token, value = self.get_token()
        self.push_token()
        if len(ans) == 1: return ans[0]
        return tuple(ans)

    def get_value(self):
        token, value = self.get_token()
        if token not in ('const', 'number', 'id', 'str'):
            self.SyntaxError("expected value, got %s" % token)
        return value

    def skip_spaces(self, pre_increment=0):
        if pre_increment:
            indent, chars = \
                scanner.count_indent(self.line[self.column:
                                               self.column+pre_increment], True)
            self.indent += indent
            self.column += chars
        indent, chars = scanner.count_indent(self.line[self.column:])
        self.indent += indent
        self.column += chars

    def parse_alternatives(self):
        r'''
            >>> from StringIO import StringIO
            >>> def do(str):
            ...    ans = StringIO(str)
            ...    ans.name = 'StringIO'
            ...    p = kqb_parser(ans)
            ...    alt, review = p.parse_alternatives()
            ...    for tag, msg in alt:
            ...        print '%s: %s' % (repr(tag), repr(msg.template))
            ...    for key, msg in sorted(review, key=lambda x: repr(x[0])):
            ...        print repr(key), '!', repr(msg.template)
            >>> do(r"""
            ...     1: hi mom
            ...        how are you?
            ...        ! Just reward!
            ...     bob: yep this is bob
            ...          ! =1
            ...     44: nope, this is just 44
            ...         ! = bob
            ... next
            ... """)
            1: u'hi mom\nhow are you?'
            'bob': u'yep this is bob'
            44: u'nope, this is just 44'
            (1, 'bob', 44) ! u'Just reward!'
        '''
        if self.column >= len(self.line):
            self.readline()
        if self.eof or self.indent == 0:
            self.SyntaxError("no alternatives", False)
        indent = self.indent
        review = {}
        ans = []
        while not self.eof and self.indent == indent:
            tag = self.get_value()
            if tag in review: self.SyntaxError("duplicate tag: %s" % tag)
            self.get_token('colon')
            ans.append((tag, Template(self.get_block_string(stop='!',
                                                            hanging=True))))
            if self.line[self.column] == '!':
                self.skip_spaces(1)
                if self.line[self.column] == '=':
                    self.indent += 1
                    self.column += 1
                    old_value = self.get_value()
                    while not isinstance(review[old_value], tuple):
                        old_value = review[old_value]
                    review[old_value][0].append(tag)
                    review[tag] = old_value
                    self.readline()
                else:
                    review[tag] = \
                        [tag], Template(self.get_block_string(hanging=True))
        if not self.eof and self.indent > indent:
            self.SyntaxError("unexpected indent", False)
        return tuple(ans), \
               tuple((value[0][0] if len(value[0]) == 1
                                  else tuple(value[0]),
                      value[1])
                     for value in review.itervalues()
                      if isinstance(value, tuple)) \
                  if review \
                  else None

    def parse_review(self):
        r'''
            >>> from StringIO import StringIO
            >>> def do(str):
            ...    ans = StringIO(str)
            ...    ans.name = 'StringIO'
            ...    p = kqb_parser(ans)
            ...    review = p.parse_review()
            ...    for key, msg in sorted(review, key=lambda x: repr(x[0])):
            ...        print repr(key), '!', repr(msg.template)
            >>> do(r"""
            ...     1 ! hi mom
            ...         how are you?
            ...         ! Just reward!
            ...     bob! yep this is bob
            ...     3-5! nope, this is just 44
            ... next
            ... """)
            'bob' ! u'yep this is bob'
            1 ! u'hi mom\nhow are you?\n! Just reward!'
            slice(3, 5, None) ! u'nope, this is just 44'
        '''
        if self.column >= len(self.line):
            self.readline()
        if self.eof or self.indent == 0:
            #print "parse_review: None"  # FIX
            return None
        #print "parse_review: self.indent", self.indent, \
        #      "self.column", self.column   # FIX
        indent = self.indent
        review = []
        while not self.eof and self.indent == indent:
            match = self.parse_match()
            self.get_token('bang')
            review.append(
                (match, Template(self.get_block_string(hanging=True))))
        if not self.eof and self.indent > indent:
            self.SyntaxError("unexpected indent", False)
        #print "parse_review:", tuple(review)   # FIX
        return tuple(review)

    def parse_questions(self):
        r''' question_base.question generator.

            >>> from StringIO import StringIO
            >>> def do(str):
            ...    ans = StringIO(str)
            ...    ans.name = 'StringIO'
            ...    p = kqb_parser(ans)
            ...    for q in p.parse_questions():
            ...        print q
            >>> do(r"""
            ... question1($ans)
            ...     This is the question?
            ...     ---
            ...     $ans = yn
            ...
            ... question2($ans)
            ...     This is the second question?
            ...     ---
            ...     $ans = select_1
            ...         1: first
            ...         2: second
            ...         3: third
            ...
            ... """)
            <question question1($ans): $ans = <yn: u'This is the question?'>>
            <question question2($ans): $ans = <select_1(1: 2: 3:): u'This is the second question?'>>
        '''
        self.readline()
        while not self.eof:
            if self.indent > 0: self.SyntaxError("unexpected indent", False)
            token, name = self.get_token('id')
            self.get_token('lparen')
            params = []
            token, param = self.get_token()
            if token != 'rparen':
                while True:
                    if token != 'param':
                        self.SyntaxError("expected $param, got %s" % token)
                    params.append(param)
                    token, ignore = self.get_token()
                    if token == 'rparen': break
                    if token != 'comma':
                        self.SyntaxError("expected comma or rparen, got %s" %
                                         token)
                    token, param = self.get_token()
            format = self.get_block_string(stop='---', ending_newlines=True)
            self.readline()     # ---
            token, answer_param = self.get_token('param')
            self.get_token('equal')
            token, cls = self.get_token('id')
            user_q = getattr(user_question, cls)(format)
            user_q.parse(self)
            yield question_base.question(name, tuple(params), answer_param,
                                         user_q)
            if self.column >= len(self.line): self.readline()

def parse_kqb(filename):
    name = os.path.basename(filename)[:-4]
    with open(filename, 'rU') as f:
        base = question_base.question_base(name)
        parser = kqb_parser(f)
        for question in parser.parse_questions():
            base.add_question(question)
    return base




# $Id: special.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import subprocess
import contextlib
from pyke import knowledge_base, rule_base

# claim_goal, fact, prove_all, gather_all

class special_knowledge_base(knowledge_base.knowledge_base):
    def __init__(self, engine):
        super(special_knowledge_base, self).__init__(engine, 'special')

    def add_fn(self, fn):
        if fn.name in self.entity_lists:
            raise KeyError("%s.%s already exists" % (self.name, fn.name))
        self.entity_lists[fn.name] = fn

    def print_stats(self, f):
        pass

class special_fn(knowledge_base.knowledge_entity_list):
    def __init__(self, special_base, name):
        super(special_fn, self).__init__(name)
        special_base.add_fn(self)

    def lookup(self, bindings, pat_context, patterns):
        raise AssertionError("special.%s may not be used in forward chaining "
                             "rules" % self.name)

    def prove(self, bindings, pat_context, patterns):
        raise AssertionError("special.%s may not be used in backward chaining "
                             "rules" % self.name)

class special_both(special_fn):
    def prove(self, bindings, pat_context, patterns):
        return self.lookup(bindings, pat_context, patterns)

class claim_goal(special_fn):
    r'''
        >>> class stub(object):
        ...     def add_fn(self, fn): pass
        >>> cg = claim_goal(stub())
        >>> mgr = cg.prove(None, None, None)
        >>> gen = iter(mgr.__enter__())
        >>> gen.next()
        >>> gen.next()
        Traceback (most recent call last):
            ...
        StopProof
        >>> mgr.__exit__(None, None, None)
        >>> cg.lookup(None, None, None)
        Traceback (most recent call last):
            ...
        AssertionError: special.claim_goal may not be used in forward chaining rules
    '''
    def __init__(self, special_base):
        super(claim_goal, self).__init__(special_base, 'claim_goal')

    def prove(self, bindings, pat_context, patterns):
        def gen():
            yield
            raise rule_base.StopProof

        return contextlib.closing(gen())

def run_cmd(pat_context, cmd_pat, cwd_pat=None, stdin_pat=None):
    r'''
        >>> from pyke import pattern
        >>> run_cmd(None, pattern.pattern_literal(('true',)))
        (0, '', '')
        >>> run_cmd(None, pattern.pattern_literal(('false',)))
        (1, '', '')
        >>> ret, out, err = run_cmd(None, pattern.pattern_literal(('pwd',)))
        >>> ret
        0
        >>> err
        ''
        >>> import os
        >>> cwd = os.getcwd() + '\n'
        >>> out == cwd
        True
        >>> run_cmd(None, pattern.pattern_literal(('pwd',)),
        ...         pattern.pattern_literal('/home/bruce'))
        (0, '/home/bruce\n', '')
    '''
    stdin = None if stdin_pat is None \
                 else stdin_pat.as_data(pat_context)
    process = subprocess.Popen(cmd_pat.as_data(pat_context),
                               bufsize=-1,
                               universal_newlines=True,
                               stdin=subprocess.PIPE,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE,
                               cwd= None if cwd_pat is None
                                         else cwd_pat.as_data(pat_context))
    out, err = process.communicate(stdin)
    return process.returncode, out, err

class check_command(special_both):
    r'''
        >>> from pyke import pattern, contexts
        >>> class stub(object):
        ...     def add_fn(self, fn): pass
        >>> cc = check_command(stub())
        >>> ctxt = contexts.simple_context()
        >>> mgr = cc.lookup(ctxt, ctxt, (pattern.pattern_literal(('true',)),))
        >>> gen = iter(mgr.__enter__())
        >>> gen.next()
        >>> ctxt.dump()
        >>> gen.next()
        Traceback (most recent call last):
            ...
        StopIteration
        >>> ctxt.dump()
        >>> mgr.__exit__(None, None, None)
        >>> mgr = cc.lookup(ctxt, ctxt, (pattern.pattern_literal(('false',)),))
        >>> gen = iter(mgr.__enter__())
        >>> gen.next()
        Traceback (most recent call last):
            ...
        StopIteration
        >>> ctxt.dump()
        >>> mgr.__exit__(None, None, None)
        >>> mgr = cc.prove(ctxt, ctxt, (pattern.pattern_literal(('true',)),))
        >>> gen = iter(mgr.__enter__())
        >>> gen.next()
        >>> ctxt.dump()
        >>> gen.next()
        Traceback (most recent call last):
            ...
        StopIteration
        >>> ctxt.dump()
        >>> mgr.__exit__(None, None, None)
    '''
    def __init__(self, special_base):
        super(check_command, self).__init__(special_base, 'check_command')

    def lookup(self, bindings, pat_context, patterns):
        if len(patterns) < 1: return knowledge_base.Gen_empty
        retcode, out, err = run_cmd(pat_context, patterns[0],
                                    patterns[1] if len(patterns) > 1
                                                else None,
                                    patterns[2] if len(patterns) > 2
                                                else None)
        if retcode: return knowledge_base.Gen_empty
        return knowledge_base.Gen_once

class command(special_both):
    r'''
        >>> from pyke import pattern, contexts
        >>> class stub(object):
        ...     def add_fn(self, fn): pass
        >>> c = command(stub())
        >>> ctxt = contexts.simple_context()
        >>> mgr = c.lookup(ctxt, ctxt,
        ...                (contexts.variable('ans'),
        ...                 pattern.pattern_literal(('echo', 'hi'))))
        >>> gen = iter(mgr.__enter__())
        >>> gen.next()
        >>> ctxt.dump()
        ans: ('hi',)
        >>> gen.next()
        Traceback (most recent call last):
            ...
        StopIteration
        >>> ctxt.dump()
        >>> mgr.__exit__(None, None, None)
        >>> mgr = c.lookup(ctxt, ctxt,
        ...                (contexts.variable('ans'),
        ...                 pattern.pattern_literal(('cat',)),
        ...                 pattern.pattern_literal(None),
        ...                 pattern.pattern_literal('line1\nline2\nline3\n')))
        >>> gen = iter(mgr.__enter__())
        >>> gen.next()
        >>> ctxt.dump()
        ans: ('line1', 'line2', 'line3')
        >>> gen.next()
        Traceback (most recent call last):
            ...
        StopIteration
        >>> ctxt.dump()
        >>> mgr.__exit__(None, None, None)
    '''
    def __init__(self, special_base):
        super(command, self).__init__(special_base, 'command')

    def lookup(self, bindings, pat_context, patterns):
        if len(patterns) < 2: return knowledge_base.Gen_empty
        retcode, out, err = run_cmd(pat_context, patterns[1],
                                    patterns[2] if len(patterns) > 2
                                                else None,
                                    patterns[3] if len(patterns) > 3
                                                else None)
        if retcode != 0:
            raise subprocess.CalledProcessError(
                                retcode,
                                ' '.join(patterns[1].as_data(pat_context)))
        def gen():
            mark = bindings.mark(True)
            try:
                outlines = tuple(out.rstrip('\n').split('\n'))
                if patterns[0].match_data(bindings, pat_context, outlines):
                    bindings.end_save_all_undo()
                    yield
                else:
                    bindings.end_save_all_undo()
            finally:
                bindings.undo_to_mark(mark)

        return contextlib.closing(gen())

class general_command(special_both):
    r'''
        >>> from pyke import pattern, contexts
        >>> class stub(object):
        ...     def add_fn(self, fn): pass
        >>> gc = general_command(stub())
        >>> ctxt = contexts.simple_context()
        >>> ctxt.dump()
        >>> mgr = gc.lookup(ctxt, ctxt,
        ...                 (contexts.variable('ans'),
        ...                  pattern.pattern_literal(('echo', 'hi'))))
        >>> gen = iter(mgr.__enter__())
        >>> gen.next()
        >>> ctxt.dump()
        ans: (0, 'hi\n', '')
        >>> gen.next()
        Traceback (most recent call last):
            ...
        StopIteration
        >>> ctxt.dump()
        >>> mgr.__exit__(None, None, None)
    '''
    def __init__(self, special_base):
        super(general_command, self).__init__(special_base, 'general_command')

    def lookup(self, bindings, pat_context, patterns):
        if len(patterns) < 2: return knowledge_base.Gen_empty
        ans = run_cmd(pat_context, patterns[1],
                      patterns[2] if len(patterns) > 2 else None,
                      patterns[3] if len(patterns) > 3 else None)

        def gen():
            mark = bindings.mark(True)
            try:
                if patterns[0].match_data(bindings, pat_context, ans):
                    bindings.end_save_all_undo()
                    yield
                else:
                    bindings.end_save_all_undo()
            finally:
                bindings.undo_to_mark(mark)

        return contextlib.closing(gen())

def create_for(engine):
    special_base = special_knowledge_base(engine)
    claim_goal(special_base)
    check_command(special_base)
    command(special_base)
    general_command(special_base)




# $Id: contexts.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
    A context is used to store all of a rule's pattern variable bindings for a
    specific rule invocation during inferencing (both forward-chaining and
    backward-chaining rules).

        >>> from pyke import pattern
        >>> c = simple_context()
        >>> var = variable('foo')
        >>> var
        $foo
        >>> var.name
        'foo'
        >>> c.bind(var.name, c, 123)
        True
        >>> c.lookup(var)
        (123, None)
        >>> c.lookup_data(var.name)
        123
    
    This is a shallow binding scheme which must "unbind" variables when a
    rule is done.  Each time a rule is invoked, a new context object is
    created to store the bindings for the variables in that rule.  When the
    rule is done, this context is abandoned; so the variables bound there
    do not need to be individually unbound.  But bindings done to
    variables in other contexts _do_ need to be individually unbound.
    These bindings happen when a variable in the calling rule, which must
    be bound in that rule's context, is matched to a non-variable pattern (i.e.,
    pattern_literal or pattern_tuple) in the called rule's context.

    For example, a caller, rule A, has its own context:

        >>> A_context = simple_context()

    and a pattern of $foo within a subgoal in its 'when' clause:

        >>> A_pattern = variable('foo')

    In proving this subgoal, rule B is being tried.  A new context is
    created for rule B:

        >>> B_context = simple_context()

    Rule B has a literal 123 in its 'proven' clause:

        >>> B_pattern = pattern.pattern_literal(123)

    Now B's 'proven' clause is pattern matched to A's subgoal, which
    results in:

        >>> B_pattern.match_pattern(B_context, B_context, A_pattern, A_context)
        True

    The pattern matches!  But the $foo variable belongs to A, so it must be
    bound in A's context, not B's.

        >>> A_context.lookup(A_pattern)
        (123, None)
        >>> B_context.lookup(A_pattern)
        Traceback (most recent call last):
            ...
        KeyError: '$foo not bound'

    This is done by using the current rule's context as a controlling context
    for all the bindings that occur within that rule.  If the binding is for
    a variable in another context, the controlling context remembers the
    binding in its undo_list.  When the rule is finished, it executes a 'done'
    method on its context which undoes all of the bindings in its undo_list.

        >>> B_context.done()
        >>> A_context.lookup(A_pattern)
        Traceback (most recent call last):
            ...
        KeyError: '$foo not bound'

    Thus, in binding a variable to a value, there are three contexts
    involved:

        controlling_context.bind(variable, variable_context,
                                 value [, value_context])

    The value_context must be omitted if value is any python data, and
    must be included if value is a pattern (including another variable).

    Variables are bound to as just strings so that they can be accessed
    easily from python code without needing the variable objects.

        >>> A_context.bind('foo', A_context, 123)
        True
        >>> A_context.lookup_data('foo')
        123
    
    But to differentiate variables from string literals in the patterns,
    variables are special objects.  When a variable is bound as data (to
    another variable), it is bound as a variable object (again, to
    differentiate it from plain python strings).

        >>> a_var = variable('a_var')
        >>> a_var
        $a_var
        >>> b_var = variable('b_var')
        >>> b_var
        $b_var
        >>> B_context.bind(b_var.name, B_context, a_var, A_context)
        True
        >>> B_context.lookup(b_var)
        Traceback (most recent call last):
            ...
        KeyError: '$a_var not bound'
        >>> ans = B_context.lookup(b_var, True)
        >>> ans                                 # doctest: +ELLIPSIS
        ($a_var, <pyke.contexts.simple_context object at ...>)
        >>> ans[1] is A_context
        True
    
    But to differentiate variables from string literals in the patterns,
    variables are special objects.  When a variable is bound as data (to
    another variable), it is bound as a variable object (again, to
    differentiate it from plain python strings).

        >>> type(ans[0])
        <class 'pyke.contexts.variable'>

    The anonymous variables have names starting with '_'.  Binding
    requests on anonymous variables are silently ignored.

        >>> anonymous('_ignored')
        $_ignored
        >>> A_context.bind('_bogus', A_context, 567)
        False
        >>> A_context.lookup_data('_bogus')
        Traceback (most recent call last):
            ...
        KeyError: '$_bogus not bound'
        >>> A_context.lookup_data('_bogus', True)
        '$_bogus'
"""

import sys

from pyke import pattern, unique

_Not_found = unique.unique('Not_found')

# Set to a sequence (or frozenset) of variable names to trace their bindings:
debug = ()

class simple_context(object):
    def __init__(self):
        self.bindings = {}
        self.undo_list = []
        self.save_all_undo_count = 0

    def dump(self):
        for var_name in sorted(self.bindings.iterkeys()):
            print "%s: %s" % (var_name, repr(self.lookup_data(var_name, True)))

    def bind(self, var_name, var_context, val, val_context = None):
        """ val_context must be None iff val is not a pattern.
            Returns True if a new binding was created.
        """
        assert not isinstance(val, pattern.pattern) \
                   if val_context is None \
                   else isinstance(val, pattern.pattern)
        if var_name[0] == '_': return False
        if var_context is self:
            assert var_name not in self.bindings
            if val_context is not None:
                val, val_context = val_context.lookup(val, True)
            if val_context == var_context and isinstance(val, variable) and \
               val.name == var_name:
                # binding $x to $x; no binding necessary!
                return False
            if var_name in debug:
                if val_context:
                    sys.stderr.write("binding %s in %s to %s in %s\n" %
                        (var_name, var_context, val, val_context))
                else:
                    sys.stderr.write("binding %s in %s to %s\n" %
                        (var_name, var_context, val))
            self.bindings[var_name] = (val, val_context)
            if self.save_all_undo_count:
                self.undo_list.append((var_name, self))
            return True
        ans = var_context.bind(var_name, var_context, val, val_context)
        if ans: self.undo_list.append((var_name, var_context))
        return ans

    def is_bound(self, var):
        val, where = var, self
        while where is not None and isinstance(val, variable):
            ans = where.bindings.get(val.name)
            if ans is None: return False
            val, where = ans
        # where is None or not isinstance(val, variable)
        return where is None or val.is_data(where)

    def lookup_data(self, var_name, allow_vars = False, final = None):
        """ Converts the answer into data only (without any patterns in it).
            If there are unbound variables anywhere in the data, a KeyError is
            generated.
        """
        if final is not None:
            val = final.get((var_name, self), _Not_found)
            if val is not _Not_found: return val
        binding = self.bindings.get(var_name)
        if binding is None:
            if allow_vars: return "$" + var_name
            raise KeyError("$%s not bound" % var_name)
        val, context = binding
        if context is not None:
            val = val.as_data(context, allow_vars, final)
        if isinstance(val, bc_context): val = val.create_plan(final)
        if final is not None: final[var_name, self] = val
        return val

    def lookup(self, var, allow_variable_in_ans = False):
        """ Returns value, val_context.
            Returns (var, self) if not bound and allow_variable_in_ans, else
            raises KeyError.
        """
        val, where = var, self
        while where is not None and isinstance(val, variable):
            ans = where.bindings.get(val.name)
            if ans is None: break
            val, where = ans
        else:
            # where is None or not isinstance(val, variable)
            return val, where
        # where is not None and isinstance(val, variable)
        if allow_variable_in_ans: return val, where
        raise KeyError("%s not bound" % str(val))

    def mark(self, save_all_undo = False):
        if save_all_undo: self.save_all_undo_count += 1
        return len(self.undo_list)

    def end_save_all_undo(self):
        assert self.save_all_undo_count > 0
        self.save_all_undo_count -= 1

    def undo_to_mark(self, mark, *var_names_to_undo):
        for var_name, var_context in self.undo_list[mark:]:
            var_context._unbind(var_name)
        del self.undo_list[mark:]
        for var_name in var_names_to_undo:
            self._unbind(var_name)

    def done(self):
        """ Unbinds all variables bound through 'self's 'bind' method.
            The assumption here is that 'self' is being abandoned, so we don't
            need to worry about self.bindings.
        """
        for var_name, var_context in self.undo_list:
            var_context._unbind(var_name)

    def _unbind(self, var_name):
        del self.bindings[var_name]

class bc_context(simple_context):
    def __init__(self, rule):
        super(bc_context, self).__init__()
        self.rule = rule

    def name(self): return self.rule.name

    def __repr__(self):
        return "<bc_context for %s at 0x%x>" % (self.name(), id(self))

    def create_plan(self, final = None):
        if final is None: final = {}
        return self.rule.make_plan(self, final)

class variable(pattern.pattern):
    """ The code to force variables of the same name to be the same object is
        probably not needed anymore...
    """
    Variables = {}

    def __new__(cls, name):
        var = cls.Variables.get(name)
        if var is None:
            var = super(variable, cls).__new__(cls)
            cls.Variables[name] = var
        return var

    def __init__(self, name):
        self.name = name

    def __repr__(self): return '$' + self.name

    def lookup(self, my_context, allow_variable_in_ans = False):
        return my_context.lookup(self, allow_variable_in_ans)

    def match_data(self, bindings, my_context, data):
        if self.name in debug:
            sys.stderr.write("%s.match_data(%s, %s, %s)\n" %
                (self, bindings, my_context, data))
        var, var_context = my_context.lookup(self, True)
        if isinstance(var, variable):
            bindings.bind(var.name, var_context, data)
            return True
        if self.name in debug:
            sys.stderr.write("%s.match_data: lookup got %s in %s\n" %
                (self, var, var_context))
        if var_context is None: return var == data
        return var.match_data(bindings, var_context, data)

    def simple_match_pattern(self, bindings, my_context, pattern_b, b_context):
        var, var_context = my_context.lookup(self, True)
        if isinstance(var, variable):
            bindings.bind(var.name, var_context, pattern_b, b_context)
            return True
        if var_context is None:
            return pattern_b.match_data(bindings, b_context, var)
        return var.simple_match_pattern(bindings, var_context,
                                        pattern_b, b_context)

    def match_pattern(self, bindings, my_context, pattern_b, b_context):
        var, var_context = my_context.lookup(self, True)
        if isinstance(var, variable):
            bindings.bind(var.name, var_context, pattern_b, b_context)
            return True
        if var_context is None:
            return pattern_b.match_data(bindings, b_context, var)
        return var.match_pattern(bindings, var_context, pattern_b, b_context)

    def as_data(self, my_context, allow_vars = False, final = None):
        return my_context.lookup_data(self.name, allow_vars, final)

    def is_data(self, my_context):
        return my_context.is_bound(self)

class anonymous(variable):
    def __init__(self, name):
        assert name[0] == '_', \
               "anonymous variables must start with '_', not %s" % name
        super(anonymous, self).__init__(name)

    def lookup(self, my_context, allow_variable_in_ans = False):
        if allow_variable_in_ans: return self, my_context
        raise KeyError("$%s not bound" % self.name)

    def match_data(self, bindings, my_context, data):
        return True

    def match_pattern(self, bindings, my_context, pattern_b, b_context):
        return True

    def as_data(self, my_context, allow_vars = False, final = None):
        if allow_vars: return "$%s" % self.name
        raise KeyError("$%s not bound" % self.name)

    def is_data(self, my_context):
        return False




# $Id: fact_base.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

'''
    A fact_base is one of the kinds of knowledge_bases (see also, rule_base
    and special).

        >>> from pyke import knowledge_engine
        >>> engine = knowledge_engine.engine()
        >>> fb = fact_base(engine, 'fb_name')
        >>> fb
        <fact_base fb_name>
        >>> fb.dump_universal_facts()
        >>> fb.dump_specific_facts()

    A fact_base is nothing more than a list of facts.  Each fact has a name
    and a tuple of arguments.  These arguments are python data (not
    patterns).

    Fact_bases support two kinds of facts: universal facts (universally
    true) and case specific facts (only true in a specific situation).

        >>> fb.add_universal_fact('some_universal_fact', ('a', 2))
        >>> fb.add_case_specific_fact('some_specific_fact', ('b', ('hi', 32)))
        >>> fb.dump_universal_facts()
        some_universal_fact('a', 2)
        >>> fb.dump_specific_facts()
        some_specific_fact('b', ('hi', 32))

    The 'reset' method deletes all case specific facts, but leaves the
    universal facts.

        >>> fb.reset()
        >>> fb.dump_universal_facts()
        some_universal_fact('a', 2)
        >>> fb.dump_specific_facts()

    Normally, universal facts are established once at program
    initialization time and case specific facts are established both just
    prior to each invocation of the expert system as well as by assertions
    in forward chaining rules.

        >>> fb.assert_('some_fact', ('a', 2, ('hi', 'mom')))
        >>> fb.dump_universal_facts()
        some_universal_fact('a', 2)
        >>> fb.dump_specific_facts()
        some_fact('a', 2, ('hi', 'mom'))
        >>> fb.assert_('some_fact', ('a', 3, ('hi', 'mom')))
        >>> fb.dump_specific_facts()
        some_fact('a', 2, ('hi', 'mom'))
        some_fact('a', 3, ('hi', 'mom'))
        >>> fb.assert_('some_other_fact', ())
        >>> fb.dump_specific_facts()
        some_fact('a', 2, ('hi', 'mom'))
        some_fact('a', 3, ('hi', 'mom'))
        some_other_fact()
    
    Duplicate facts are not allowed and trying to assert a duplicate fact is
    silently ignored.

        >>> fb.assert_('some_fact', ('a', 2, ('hi', 'mom')))
        >>> fb.dump_specific_facts()
        some_fact('a', 2, ('hi', 'mom'))
        some_fact('a', 3, ('hi', 'mom'))
        some_other_fact()

'''

import itertools
import contextlib
from pyke import knowledge_base, contexts

class fact_base(knowledge_base.knowledge_base):
    ''' Not much to fact_bases.  The real work is done in fact_list! '''
    def __init__(self, engine, name, register = True):
        super(fact_base, self).__init__(engine, name, fact_list, register)

    def dump_universal_facts(self):
        for fl_name in sorted(self.entity_lists.iterkeys()):
            self.entity_lists[fl_name].dump_universal_facts()

    def dump_specific_facts(self):
        for fl_name in sorted(self.entity_lists.iterkeys()):
            self.entity_lists[fl_name].dump_specific_facts()

    def add_universal_fact(self, fact_name, args):
        self.get_entity_list(fact_name).add_universal_fact(args)

    def add_case_specific_fact(self, fact_name, args):
        self.get_entity_list(fact_name).add_case_specific_fact(args)

    def assert_(self, fact_name, args):
        self.add_case_specific_fact(fact_name, args)

    def get_stats(self):
        num_fact_lists = num_universal = num_case_specific = 0
        for fact_list in self.entity_lists.itervalues():
            universal, case_specific = fact_list.get_stats()
            num_universal += universal
            num_case_specific += case_specific
            num_fact_lists += 1
        return num_fact_lists, num_universal, num_case_specific

    def print_stats(self, f):
        num_fact_lists, num_universal, num_case_specific = self.get_stats()
        f.write("%s: %d fact names, %d universal facts, "
                "%d case_specific facts\n" %
                (self.name, num_fact_lists, num_universal, num_case_specific))

class fact_list(knowledge_base.knowledge_entity_list):
    def __init__(self, name):
        super(fact_list, self).__init__(name)
        self.universal_facts = []               # [(arg...)...]
        self.case_specific_facts = []           # [(arg...)...]
        self.hashes = {}        # (len, (index...)): (other_indices,
                                #       {(arg...): [other_args_from_factn...]})
        self.fc_rule_refs = []  # (fc_rule, foreach_index)

    def reset(self):
        self.case_specific_facts = []
        self.hashes.clear()
        self.fc_rule_refs = []

    def dump_universal_facts(self):
        for args in self.universal_facts:
            print '%s%s' % (self.name, args)

    def dump_specific_facts(self):
        for args in self.case_specific_facts:
            print '%s%s' % (self.name, args)

    def add_fc_rule_ref(self, fc_rule, foreach_index):
        self.fc_rule_refs.append((fc_rule, foreach_index))

    def get_affected_fc_rules(self):
        return (fc_rule for fc_rule, foreach_index in self.fc_rule_refs)

    def lookup(self, bindings, pat_context, patterns):
        """ Returns a context manager for a generator that binds patterns to
            successive facts, yielding None for each successful match.
            Undoes bindings upon continuation, so that no bindings remain at
            StopIteration.
        """
        indices = tuple(enum for enum in enumerate(patterns)
                             if enum[1].is_data(pat_context))
        other_indices, other_arg_lists = \
            self._get_hashed(len(patterns),
                             tuple(index[0] for index in indices),
                             tuple(index[1].as_data(pat_context)
                                   for index in indices))
        def gen():
            if other_arg_lists:
                for args in other_arg_lists:
                    mark = bindings.mark(True)
                    end_done = False
                    try:
                        if all(itertools.imap(
                                   lambda i, arg:
                                       patterns[i].match_data(bindings,
                                                              pat_context,
                                                              arg),
                                   other_indices,
                                   args)):
                            bindings.end_save_all_undo()
                            end_done = True
                            yield
                    finally:
                        if not end_done: bindings.end_save_all_undo()
                        bindings.undo_to_mark(mark)
        return contextlib.closing(gen())

    def _get_hashed(self, len, indices, args):
        ans = self.hashes.get((len, indices))
        if ans is None: ans = self._hash(len, indices)
        other_indices, arg_map = ans
        return other_indices, arg_map.get(args, ())

    def _hash(self, length, indices):
        args_hash = {}
        new_entry = (tuple(i for i in range(length) if i not in indices),
                     args_hash)
        self.hashes[length, indices] = new_entry
        for args in itertools.chain(self.universal_facts,
                                    self.case_specific_facts):
            if len(args) == length:
                selected_args = tuple(arg for i, arg in enumerate(args)
                                          if i in indices)
                args_hash.setdefault(selected_args, []) \
                         .append(tuple(arg for i, arg in enumerate(args)
                                           if i not in indices))
        return new_entry

    def add_universal_fact(self, args):
        assert args not in self.case_specific_facts, \
               "add_universal_fact: fact already present as specific fact"
        if args not in self.universal_facts:
            self.universal_facts.append(args)
            self.add_args(args)

    def add_case_specific_fact(self, args):
        if args not in self.universal_facts and \
           args not in self.case_specific_facts:
            self.case_specific_facts.append(args)
            self.add_args(args)
            for fc_rule, foreach_index in self.fc_rule_refs:
                fc_rule.new_fact(args, foreach_index)

    def add_args(self, args):
        for (length, indices), (other_indices, arg_map) \
         in self.hashes.iteritems():
            if length == len(args):
                selected_args = tuple(arg for i, arg in enumerate(args)
                                          if i in indices)
                arg_map.setdefault(selected_args, []) \
                       .append(tuple(arg for i, arg in enumerate(args)
                                         if i not in indices))

    def get_stats(self):
        return len(self.universal_facts), len(self.case_specific_facts)




# $Id: krbparser.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

""" See http://www.dabeaz.com/ply/ply.html for syntax of grammer definitions.
""" 

from __future__ import with_statement
import itertools
import warnings
import os, os.path
from pyke.krb_compiler.ply import yacc
from pyke.krb_compiler import scanner
from pyke import pattern, contexts

tokens = scanner.tokens

goal_mode = False

def p_top(p):
    ''' top : file
            | python_goal
    '''
    p[0] = p[1]

def p_goal(p):
    ''' python_goal : CHECK_TOK IDENTIFIER_TOK '.' IDENTIFIER_TOK LP_TOK patterns_opt RP_TOK
    '''
    p[0] = (p[2], p[4], p[6], pattern_vars)

def p_file(p):
    ''' file : nl_opt parent_opt fc_rules bc_rules_opt
    '''
    p[0] = ('file', p[2], (tuple(p[3]), ()), p[4])

def p_file_fc_extras(p):
    ''' file : nl_opt parent_opt fc_rules fc_extras bc_rules_opt
    '''
    p[0] = ('file', p[2], (tuple(p[3]), p[4]), p[5])

def p_file_bc(p):
    ''' file : nl_opt parent_opt bc_rules_section
    '''
    p[0] = ('file', p[2], ((), ()), p[3])

# Uncomment this to generate an error in the grammer.
#def p_bogus(p):
#    ''' file : nl_opt parent_opt bc_rules_opt
#    '''
#    p[0] = ('file', p[2], ((), ()), p[3])

def p_parent(p):
    ''' parent_opt : EXTENDING_TOK IDENTIFIER_TOK without_opt NL_TOK
    '''
    p[0] = ('parent', p[2], tuple(p[3]))

def p_second(p):
    ''' without_opt : WITHOUT_TOK without_names comma_opt
    '''
    p[0] = p[2]

def p_fourth(p):
    ''' when_opt : WHEN_TOK NL_TOK reset_plan_vars INDENT_TOK bc_premises DEINDENT_TOK
    '''
    p[0] = p[5]

def p_reset_plan_vars(p):
    ''' reset_plan_vars :
    '''
    global plan_var, plan_var_number
    plan_var_number = 1
    plan_var = 'plan#%d' % plan_var_number
    p[0] = None

def p_inc_plan_vars(p):
    ''' inc_plan_vars :
    '''
    global plan_var, plan_var_number
    plan_var_number += 1
    plan_var = 'plan#%d' % plan_var_number
    p[0] = None

def p_fifth(p):
    ''' bc_extras_opt : BC_EXTRAS_TOK NL_TOK start_extra_statements INDENT_TOK python_extras_code NL_TOK DEINDENT_TOK
        fc_extras : FC_EXTRAS_TOK NL_TOK start_extra_statements INDENT_TOK python_extras_code NL_TOK DEINDENT_TOK
        plan_extras_opt : PLAN_EXTRAS_TOK NL_TOK start_extra_statements INDENT_TOK python_extras_code NL_TOK DEINDENT_TOK
        with_opt : WITH_TOK NL_TOK start_python_statements INDENT_TOK python_plan_code NL_TOK DEINDENT_TOK
    '''
    p[0] = p[5]

def p_none(p):
    ''' bc_require_opt :
        comma_opt :
        comma_opt : ','
        colon_opt :
        data : NONE_TOK
        fc_require_opt :
        nl_opt :
        nl_opt : NL_TOK
        parent_opt :
        plan_spec : NL_TOK
        rest_opt : comma_opt
    '''
    p[0] = None

def p_colon_deprication(p):
    ''' colon_opt : ':'
    '''
    warnings.warn_explicit("use of ':' deprecated after rule names",
                           DeprecationWarning,
                           scanner.lexer.filename, p.lineno(1))
    p[0] = None

def p_fc_rule(p):
    ''' fc_rule : IDENTIFIER_TOK colon_opt NL_TOK INDENT_TOK foreach_opt ASSERT_TOK NL_TOK INDENT_TOK assertions DEINDENT_TOK DEINDENT_TOK
    '''
    p[0] = ('fc_rule', p[1], p[5], tuple(p[9]))

def p_foreach(p):
    ''' foreach_opt : FOREACH_TOK NL_TOK INDENT_TOK fc_premises DEINDENT_TOK
    '''
    p[0] = tuple(p[4])

def p_fc_premise(p):
    ''' fc_premise : IDENTIFIER_TOK '.' IDENTIFIER_TOK LP_TOK patterns_opt RP_TOK NL_TOK
    '''
    p[0] = ('fc_premise', p[1], p[3], tuple(p[5]), p.lineno(1), p.lineno(6))

def p_fc_first_1(p):
    ''' fc_premise : FIRST_TOK NL_TOK INDENT_TOK fc_premises DEINDENT_TOK
    '''
    p[0] = ('fc_first', tuple(p[4]), p.lineno(1))

def p_fc_first_n(p):
    ''' fc_premise : FIRST_TOK fc_premise
    '''
    p[0] = ('fc_first', (p[2],), p.lineno(1))

def p_fc_notany(p):
    ''' fc_premise : NOTANY_TOK NL_TOK INDENT_TOK fc_premises DEINDENT_TOK
    '''
    p[0] = ('fc_notany', tuple(p[4]), p.lineno(1))

def p_fc_forall(p):
    ''' fc_premise : FORALL_TOK NL_TOK INDENT_TOK fc_premises DEINDENT_TOK fc_require_opt
    '''
    p[0] = ('fc_forall', tuple(p[4]), p[6], p.lineno(1), p.linespan(6)[1])

def p_fc_require_opt(p):
    ''' fc_require_opt : REQUIRE_TOK NL_TOK INDENT_TOK fc_premises DEINDENT_TOK
    '''
    p[0] = tuple(p[4])

def p_python_eq(p):
    ''' python_premise : pattern start_python_code '=' python_rule_code NL_TOK
    '''
    p[0] = ('python_eq', p[1], p[4], p.linespan(1)[0], p.linespan(4)[1])

def p_python_in(p):
    ''' python_premise : pattern start_python_code IN_TOK python_rule_code NL_TOK
    '''
    p[0] = ('python_in', p[1], p[4], p.linespan(1)[0], p.linespan(4)[1])

def p_python_check(p):
    ''' python_premise : start_python_code CHECK_TOK python_rule_code NL_TOK
    '''
    p[0] = ('python_check', p[3], p.lineno(2), p.linespan(3)[1])

def p_python_block_n(p):
    ''' python_premise : check_nl PYTHON_TOK NL_TOK start_python_assertion INDENT_TOK python_rule_code NL_TOK DEINDENT_TOK
    '''
    p[0] = ('python_block', p[6], p.lineno(2), p.linespan(6)[1])

def p_python_block_1(p):
    ''' python_premise : check_nl PYTHON_TOK start_python_code NOT_NL_TOK python_rule_code NL_TOK
    '''
    p[0] = ('python_block', p[5], p.lineno(2), p.linespan(5)[1])

def p_assertion(p):
    ''' assertion : IDENTIFIER_TOK '.' IDENTIFIER_TOK LP_TOK patterns_opt RP_TOK NL_TOK
    '''
    p[0] = ('assert', p[1], p[3], tuple(p[5]), p.lineno(1), p.lineno(6))

def p_python_assertion_n(p):
    ''' assertion : check_nl PYTHON_TOK NL_TOK start_python_assertion INDENT_TOK python_rule_code NL_TOK DEINDENT_TOK
    '''
    p[0] = ('python_assertion', p[6], p.lineno(2), p.linespan(6)[1])

def p_python_assertion_1(p):
    ''' assertion : check_nl PYTHON_TOK start_python_code NOT_NL_TOK python_rule_code NL_TOK
    '''
    p[0] = ('python_assertion', p[5], p.lineno(2), p.linespan(5)[1])

def p_check_nl(p):
    ''' check_nl :
    '''
    scanner.lexer.begin('checknl')
    p[0] = None

def p_bc_rule(p):
    ''' bc_rule : IDENTIFIER_TOK colon_opt NL_TOK INDENT_TOK USE_TOK goal when_opt with_opt DEINDENT_TOK
    '''
    p[0] = ('bc_rule', p[1], p[6], tuple(p[7]), tuple(p[8][0]), tuple(p[8][1]))

def p_empty_bc_rules_opt(p):
    ''' bc_rules_opt :
    '''
    p[0] = ((), (), ())

def p_bc_rules_section(p):
    ''' bc_rules_section : bc_rules bc_extras_opt plan_extras_opt
    '''
    p[0] = (tuple(p[1]), p[2], p[3])

def p_goal_no_taking(p):
    ''' goal : IDENTIFIER_TOK LP_TOK patterns_opt RP_TOK NL_TOK
    '''
    p[0] = ('goal', p[1], tuple(p[3]), (), p.lineno(1), p.lineno(4))

def p_goal_taking(p):
    ''' goal : IDENTIFIER_TOK LP_TOK patterns_opt RP_TOK taking
    '''
    p[0] = ('goal', p[1], tuple(p[3]), p[5], p.lineno(1), p.lineno(4))

def p_name_sym(p):
    ''' name : IDENTIFIER_TOK
    '''
    p[0] = repr(p[1])

def p_name_pat_var(p):
    ''' name : PATTERN_VAR_TOK
    '''
    p[0] = "context.lookup_data(%s)" % p[1]

def p_bc_premise1(p):
    ''' bc_premise : name LP_TOK patterns_opt RP_TOK plan_spec
    '''
    p[0] = ('bc_premise', False, None, p[1], tuple(p[3]), p[5],
            p.linespan(1)[0], p.lineno(4))

def p_bc_premise2(p):
    ''' bc_premise : '!' name LP_TOK patterns_opt RP_TOK plan_spec
    '''
    p[0] = ('bc_premise', True, None, p[2], tuple(p[4]), p[6],
            p.lineno(1), p.lineno(5))

def p_bc_premise3(p):
    ''' bc_premise : name '.' name LP_TOK patterns_opt RP_TOK plan_spec
    '''
    p[0] = ('bc_premise', False, p[1], p[3], tuple(p[5]), p[7],
            p.linespan(1)[0], p.lineno(6))

def p_bc_premise4(p):
    ''' bc_premise : '!' name '.' name LP_TOK patterns_opt RP_TOK plan_spec
    '''
    p[0] = ('bc_premise', True, p[2], p[4], tuple(p[6]), p[8],
            p.lineno(1), p.lineno(7))

def p_bc_first_1f(p):
    ''' bc_premise : FIRST_TOK NL_TOK INDENT_TOK bc_premises DEINDENT_TOK
    '''
    p[0] = ('bc_first', False, tuple(p[4]), p.lineno(1))

def p_bc_first_nf(p):
    ''' bc_premise : FIRST_TOK bc_premise
    '''
    p[0] = ('bc_first', False, (p[2],), p.lineno(1))

def p_bc_first_1t(p):
    ''' bc_premise : '!' FIRST_TOK NL_TOK INDENT_TOK bc_premises DEINDENT_TOK
    '''
    p[0] = ('bc_first', True, tuple(p[4]), p.lineno(1))

def p_bc_first_nt(p):
    ''' bc_premise : '!' FIRST_TOK bc_premise
    '''
    p[0] = ('bc_first', True, (p[2],), p.lineno(1))

def p_bc_notany(p):
    ''' bc_premise : NOTANY_TOK NL_TOK INDENT_TOK bc_premises DEINDENT_TOK
    '''
    p[0] = ('bc_notany', tuple(p[4]), p.lineno(1))

def p_bc_forall(p):
    ''' bc_premise : FORALL_TOK NL_TOK INDENT_TOK bc_premises DEINDENT_TOK bc_require_opt
    '''
    p[0] = ('bc_forall', tuple(p[4]), p[6], p.lineno(1), p.linespan(6)[1])

def p_bc_require_opt(p):
    ''' bc_require_opt : REQUIRE_TOK NL_TOK INDENT_TOK bc_premises DEINDENT_TOK
    '''
    p[0] = tuple(p[4])

def p_as(p):
    ''' plan_spec : AS_TOK PATTERN_VAR_TOK NL_TOK
    '''
    p[0] = ('as', p[2][1:-1])

def p_step_code(p):
    ''' plan_spec : STEP_TOK NUMBER_TOK NL_TOK start_python_plan_call INDENT_TOK python_plan_code NL_TOK DEINDENT_TOK
    '''
    p[0] = ('plan_spec', p[2], plan_var, p[6][0], p[6][1],
            p.lineno(1), p.lexpos(1))

def p_code(p):
    ''' plan_spec : NL_TOK start_python_plan_call INDENT_TOK python_plan_code NL_TOK DEINDENT_TOK
    '''
    p[0] = ('plan_spec', None, plan_var, p[4][0], p[4][1], p[4][2], p[4][3])

def p_start_python_code(p):
    ''' start_python_code :
    '''
    scanner.start_code(var_format = "context.lookup_data('%s')")
    p[0] = None

def p_start_python_plan_call(p):
    ''' start_python_plan_call :
    '''
    scanner.start_code(plan_name = plan_var, multiline = True)
    p[0] = None

def p_start_python_statements(p):
    ''' start_python_statements :
    '''
    scanner.start_code(multiline = True)
    p[0] = None

def p_start_extra_statements(p):
    ''' start_extra_statements :
    '''
    scanner.start_code(multiline = True, var_format = None)
    p[0] = None

def p_start_python_assertion(p):
    ''' start_python_assertion :
    '''
    scanner.start_code(multiline = True,
                       var_format = "context.lookup_data('%s')")
    p[0] = None

def p_python_rule_code(p):
    ''' python_rule_code : CODE_TOK
    '''
    p[0] = p[1]

def p_python_plan_code(p):
    ''' python_plan_code : CODE_TOK
    '''
    p[0] = p[1]

def p_python_extras_code(p):
    ''' python_extras_code : CODE_TOK
    '''
    p[0] = p[1][0]

def p_pattern_var(p):
    ''' variable : PATTERN_VAR_TOK
    '''
    global pattern_vars
    if goal_mode:
        pattern_vars.append(p[1])
        p[0] = contexts.variable(p[1])
    else:
        p[0] = "contexts.variable(%s)" % p[1]

def p_anonymous_var(p):
    ''' variable : ANONYMOUS_VAR_TOK
    '''
    if goal_mode:
        p[0] = contexts.anonymous(p[1])
    else:
        p[0] = "contexts.anonymous(%s)" % p[1]

def p_first(p):
    ''' bc_premise : python_premise
        bc_rules_opt : bc_rules_section
        data : NUMBER_TOK
        fc_premise : python_premise
        pattern : pattern_proper
        pattern_proper : variable
        patterns_opt : patterns comma_opt
    '''
    p[0] = p[1]

def p_last(p):
    ''' rest_opt : ',' '*' variable
    '''
    p[0] = p[len(p)-1]

def p_data_string(p):
    ''' data : STRING_TOK
    '''
    if goal_mode:
        if p[1].startswith("'''") or p[1].startswith('"""'):
            p[0] = scanner.unescape(p[1][3:-3])
        else:
            p[0] = scanner.unescape(p[1][1:-1])
    else:
        p[0] = p[1]

def p_taking(p):
    ''' taking : start_python_code TAKING_TOK python_rule_code NL_TOK
    '''
    p[0] = p[3][0]

def p_taking2(p):
    ''' taking : NL_TOK INDENT_TOK start_python_code TAKING_TOK python_rule_code NL_TOK DEINDENT_TOK
    '''
    p[0] = p[5][0]

def p_quoted_last(p):
    ''' data : IDENTIFIER_TOK
    '''
    if goal_mode:
        p[0] = p[len(p)-1]
    else:
        p[0] = "'" + p[len(p)-1] + "'"

def p_false(p):
    ''' data : FALSE_TOK
    '''
    p[0] = False

def p_true(p):
    ''' data : TRUE_TOK
    '''
    p[0] = True

def p_start_list(p):
    ''' assertions : assertion
        bc_premises : bc_premise
        bc_rules : bc_rule
        data_list : data
        fc_premises : fc_premise
        fc_rules : fc_rule
        patterns : pattern
        patterns_proper : pattern_proper
        without_names : IDENTIFIER_TOK
    '''
    p[0] = [p[1]]

def p_empty_tuple(p):
    ''' bc_extras_opt :
        data : LP_TOK RP_TOK
        foreach_opt :
        patterns_opt :
        plan_extras_opt :
        when_opt :
        without_opt :
    '''
    p[0] = ()

def p_double_empty_tuple(p):
    ''' with_opt :
    '''
    p[0] = (), ()

def p_append_list(p):
    ''' assertions : assertions assertion
        bc_premises : bc_premises inc_plan_vars bc_premise
        bc_rules : bc_rules bc_rule
        data_list : data_list ',' data
        fc_premises : fc_premises fc_premise
        fc_rules : fc_rules fc_rule
        patterns : patterns ',' pattern
        patterns_proper : patterns_proper ',' pattern
        without_names : without_names ',' IDENTIFIER_TOK
    '''
    p[1].append(p[len(p)-1])
    p[0] = p[1]

def p_pattern(p):
    ''' pattern : data '''
    if goal_mode:
        p[0] = pattern.pattern_literal(p[1])
    else:
        p[0] = "pattern.pattern_literal(%s)" % str(p[1])

def p_pattern_tuple1(p):
    ''' pattern_proper : LP_TOK '*' variable RP_TOK '''
    if goal_mode:
        p[0] = pattern.pattern_tuple((), p[3])
    else:
        p[0] = "pattern.pattern_tuple((), %s)" % p[3]

def p_pattern_tuple2(p):
    ''' pattern_proper : LP_TOK data_list ',' '*' variable RP_TOK '''
    if goal_mode:
        p[0] = pattern.pattern_tuple(
                 tuple(pattern.pattern_literal(x) for x in p[2]),
                 p[5])
    else:
        p[0] = "pattern.pattern_tuple((%s), %s)" % \
                   (' '.join("pattern.pattern_literal(%s)," % str(x)
                             for x in p[2]),
                    p[5])

def p_pattern_tuple3(p):
    ''' pattern_proper : LP_TOK data_list ',' patterns_proper rest_opt RP_TOK '''
    if goal_mode:
        p[0] = pattern.pattern_tuple(
                 tuple(itertools.chain(
                         (pattern.pattern_literal(x) for x in p[2]),
                         p[4])),
                 p[5])
    else:
        p[0] = "pattern.pattern_tuple((%s), %s)" % \
                   (' '.join(itertools.chain(
                               ("pattern.pattern_literal(%s)," % str(x)
                                 for x in p[2]),
                               (str(x) + ',' for x in p[4]))),
                    p[5])

def p_pattern_tuple4(p):
    ''' pattern_proper : LP_TOK patterns_proper rest_opt RP_TOK '''
    if goal_mode:
        p[0] = pattern.pattern_tuple(p[2], p[3])
    else:
        p[0] = "pattern.pattern_tuple((%s), %s)" % \
                   (' '.join(str(x) + ',' for x in p[2]),
                    p[3])

def p_tuple(p):
    ''' data : LP_TOK data_list comma_opt RP_TOK '''
    if goal_mode:
        p[0] = tuple(p[2])
    else:
        p[0] = '(' + ' '.join(str(x) + ',' for x in p[2]) + ')'

def p_error(t):
    if t is None:
        raise SyntaxError("invalid syntax", scanner.syntaxerror_params())
    else:
        raise SyntaxError("invalid syntax",
                          scanner.syntaxerror_params(t.lexpos, t.lineno))

parser = None

def init(this_module, check_tables = False, debug = 0):
    global parser
    if parser is None:
        outputdir = os.path.dirname(this_module.__file__)
        if debug:
            parser = yacc.yacc(module=this_module, write_tables=0,
                               debug=debug, debugfile='krbparser.yacc.out',
                               outputdir=outputdir)
        else:
            if check_tables:
                krbparser_mtime = os.path.getmtime(this_module.__file__)
                tables_name = os.path.join(outputdir, 'krbparser_tables.py')
                try:
                    ok = os.path.getmtime(tables_name) >= krbparser_mtime
                except OSError:
                    ok = False
                if not ok:
                    #print "regenerating krbparser_tables"
                    try: os.remove(tables_name)
                    except OSError: pass
                    try: os.remove(tables_name + 'c')
                    except OSError: pass
                    try: os.remove(tables_name + 'o')
                    except OSError: pass
            parser = yacc.yacc(module=this_module, debug=0,
                               optimize=1, write_tables=1,
                               tabmodule='pyke.krb_compiler.krbparser_tables',
                               outputdir=outputdir)

# Use the first line for normal use, the second for testing changes in the
# grammer (the first line does not report grammer errors!).
def parse(this_module, filename, check_tables = False, debug = 0):
#def parse(this_module, filename, check_tables = True, debug = 1):
    global goal_mode
    init(this_module, check_tables, debug)
    with open(filename) as f:
        scanner.init(scanner, debug, check_tables)
        scanner.lexer.lineno = 1
        scanner.lexer.filename = filename
        scanner.kfb_mode = False
        scanner.goal_mode = False
        goal_mode = False
        #parser.restart()
        return parser.parse(f.read() + '\n', lexer=scanner.lexer, tracking=True,
                            debug=debug)

def parse_goal(this_module, s, check_tables = False, debug = 0):
    global goal_mode, pattern_vars
    init(this_module, check_tables, debug)
    scanner.init(scanner, debug, check_tables)
    scanner.lexer.lineno = 1
    scanner.lexer.filename = s
    scanner.kfb_mode = False
    scanner.goal_mode = True
    goal_mode = True
    pattern_vars = []
    #parser.restart()
    return parser.parse('check ' + s, lexer=scanner.lexer, tracking=True,
                        debug=debug)

def run(this_module, filename='TEST/krbparse_test.krb'):
    r""" Used for testing.

        >>> import os, os.path
        >>> from pyke.krb_compiler import krbparser
        >>> run(krbparser, os.path.join(os.path.dirname(__file__),
        ...                             'TEST/krbparse_test.krb'))
        ('file',
         None,
         ((('fc_rule',
            'name1',
            (('fc_premise',
              'a',
              'b',
              ("pattern.pattern_literal('x')", "contexts.variable('b')"),
              7,
              7),),
            (('assert', 'c', 'd', ("contexts.variable('b')",), 9, 9),)),),
          ()),
         ((('bc_rule',
            'name2',
            ('goal',
             'x',
             ('pattern.pattern_literal(1)',
              "contexts.variable('c')",
              "pattern.pattern_literal((1, 'b',))",
              "pattern.pattern_tuple((pattern.pattern_literal(1), pattern.pattern_literal('b'), contexts.variable('c'),), None)"),
             (' (a, b, c)',),
             12,
             12),
            (('bc_premise',
              False,
              "'a'",
              "'b'",
              ("pattern.pattern_literal('x')", "contexts.variable('c')"),
              ('plan_spec',
               None,
               'plan#1',
               ('line1', "line2 (context['d']) end2"),
               ('d',),
               15,
               219),
              14,
              14),
             ('bc_premise',
              False,
              None,
              "'x'",
              ('pattern.pattern_literal(1)',
               'pattern.pattern_literal(2)',
               'pattern.pattern_literal(3)'),
              ('plan_spec',
               None,
               'plan#2',
               ("some (context['plan'])(stuff) \\",
                '                and more stuff fail here'),
               ('plan',),
               18,
               280),
              17,
              17)),
            ("a (context['plan']) b",),
            ('plan',)),
           ('bc_rule',
            'name3',
            ('goal',
             'x',
             ('pattern.pattern_literal(1)',
              "contexts.variable('c')",
              "pattern.pattern_literal((1, 'b',))",
              "pattern.pattern_tuple((pattern.pattern_literal(1), pattern.pattern_literal('b'), contexts.variable('c'),), None)"),
             (),
             24,
             24),
            (('bc_premise',
              False,
              "'a'",
              "'b'",
              ("pattern.pattern_literal('x')", "contexts.variable('c')"),
              ('plan_spec',
               None,
               'plan#1',
               ('line1', "line2 (context['d']) end2"),
               ('d',),
               27,
               452),
              26,
              26),
             ('bc_premise',
              False,
              None,
              "'x'",
              ('pattern.pattern_literal(1)',
               'pattern.pattern_literal(2)',
               'pattern.pattern_literal(3)'),
              ('as', 'foo_fn'),
              29,
              29)),
            (),
            ())),
          (),
          ()))

    """
    import pprint
    pprint.pprint(parse(this_module, filename, True))




# $Id: pattern.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import types
import itertools

class pattern(object):
    def __ne__(self, b): return not (self == b)

    def simple_match_pattern(self, bindings, my_context, pattern_b, b_context):
        return self.match_pattern(bindings, my_context, pattern_b, b_context)

    def lookup(self, context, allow_variable_in_ans = False):
        return self

class pattern_literal(pattern):
    def __init__(self, literal):
        self.literal = literal

    def __hash__(self): return hash(self.literal)

    def __eq__(self, b):
        if isinstance(b, pattern_literal): return self.literal == b.literal
        return self.literal == b

    def match_data(self, bindings, my_context, data):
        return self.literal == data

    def match_pattern(self, bindings, my_context, pattern_b, b_context):
        if isinstance(pattern_b, pattern_literal):
            return self.literal == pattern_b.literal
        return pattern_b.match_data(bindings, b_context, self.literal)

    def as_data(self, my_context, allow_vars = False, final = None):
        return self.literal

    def is_data(self, my_context):
        return True

class pattern_tuple(pattern):
    def __init__(self, elements, rest_var = None):
        self.elements = tuple(elements)
        self.rest_var = rest_var

    def __hash__(self):
        return hash(self.elements) ^ hash(self.rest_var)

    def __eq__(self, b):
        return isinstance(b, pattern_tuple) and \
               self.elements == b.elements and self.rest_var == b.rest_var

    def match_data(self, bindings, my_context, data):
        if isinstance(data, types.StringTypes): return False
        try:
            data = tuple(data)
        except TypeError:
            return False
        if len(self.elements) > len(data) or \
           self.rest_var is None and len(self.elements) < len(data):
            return False
        for x, y in itertools.izip(self.elements, data):
            if not x.match_data(bindings, my_context, y): return False
        if self.rest_var is not None:
            return self.rest_var.match_data(bindings, my_context,
                                            tuple(data[len(self.elements):]))
        return True

    def simple_match_pattern(self, bindings, my_context, pattern_b, b_context):
        return self, my_context

    def match_pattern(self, bindings, my_context, pattern_b, b_context):
        simple_ans = pattern_b.simple_match_pattern(bindings, b_context,
                                                    self, my_context)
        if isinstance(simple_ans, bool): return simple_ans
        pattern_b, b_context = simple_ans
        if not isinstance(pattern_b, pattern):
            return self.match_data(bindings, my_context, pattern_b)
        assert isinstance(pattern_b, pattern_tuple), "Internal logic error"

        my_len = len(self.elements)
        b_len = len(pattern_b.elements)
        if pattern_b.rest_var is None and my_len > b_len or \
           self.rest_var is None and my_len < b_len:
            return False
        for x, y in itertools.izip(self.elements, pattern_b.elements):
            if not x.match_pattern(bindings, my_context, y, b_context):
                return False
        if my_len <= b_len and self.rest_var is not None:
            # This is where the two rest_vars are bound together if my_len ==
            # b_len.
            tail_val, tail_context = pattern_b._tail(my_len, b_context)
            if tail_context is None:
                if not self.rest_var.match_data(bindings, my_context, tail_val):
                    return False
            else:
                if not self.rest_var.match_pattern(bindings, my_context,
                                                   tail_val, tail_context):
                    return False
        elif pattern_b.rest_var is not None:
            tail_val, tail_context = self._tail(b_len, my_context)
            if tail_context is None:
                if not pattern_b.rest_var.match_data(bindings, b_context,
                                                     tail_val):
                    return False
            else:
                if not pattern_b.rest_var.match_pattern(bindings, b_context,
                                                        tail_val, tail_context):
                    return False
        return True

    def as_data(self, my_context, allow_vars = False, final = None):
        ans = tuple(x.as_data(my_context, allow_vars, final)
                    for x in self.elements)
        if self.rest_var is None:
            return ans
        rest = my_context.lookup_data(self.rest_var.name, allow_vars, final)
        if isinstance(rest, tuple): return ans + rest
        return ans + ('*' + rest,)

    def _tail(self, n, my_context):
        """ Return a copy of myself with the first n elements removed.
        """
        if n == len(self.elements):
            if self.rest_var is None: return (), None
            return self.rest_var, my_context
        rest_elements = self.elements[n:]
        if self.rest_var is None and \
           all(isinstance(x, pattern_literal) for x in rest_elements):
            return tuple(x.literal for x in rest_elements), None
        return pattern_tuple(self.elements[n:], self.rest_var), my_context

    def is_data(self, my_context):
        arg_test = all(arg_pat.is_data(my_context) for arg_pat in self.elements)
        if not arg_test or self.rest_var is None: return arg_test
        return self.rest_var.is_data(my_context)




# $Id: helpers.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pyke

def fc_head(rb_name):
    return (
        "# %s_fc.py" % rb_name,
        "",
        "from __future__ import with_statement",
        "from pyke import contexts, pattern, fc_rule, knowledge_base",
        "",
        "pyke_version = %r" % pyke.version,
        "compiler_version = %r" % pyke.compiler_version,
    )

def bc_head(rb_name):
    return (
        "# %s_bc.py" % rb_name,
        "",
        "from __future__ import with_statement",
        "import itertools",
        "from pyke import contexts, pattern, bc_rule",
        "",
        "pyke_version = %r" % pyke.version,
        "compiler_version = %r" % pyke.compiler_version,
    )

def plan_head(rb_name):
    return (
        "# %s_plans.py" % rb_name,
        "",
        "pyke_version = %r" % pyke.version,
        "compiler_version = %r" % pyke.compiler_version,
    )

def goal(rb_name, rule_name, goal_info, pred_plan_lines, python_lines):
    # returns plan_lines, goal_fn_head, goal_fn_tail, goal_decl_lines
    goal, goal_name, pattern_args, taking, start_lineno, end_lineno = goal_info
    assert goal == 'goal'
    goal_fn_head = (
        "",
        "def %s(rule, arg_patterns, arg_context):" % rule_name,
        ("INDENT", 2),
        "engine = rule.rule_base.engine",
        "patterns = rule.goal_arg_patterns()",
        "if len(arg_patterns) == len(patterns):",
        ("INDENT", 2),
        "context = contexts.bc_context(rule)",
        "try:",
        ("INDENT", 2),
        ("STARTING_LINENO", start_lineno),
        "if all(itertools.imap(lambda pat, arg:",
        ("INDENT", 2),
        ("INDENT", 20),
        ("INDENT", 2),
        "pat.match_pattern(context, context,",
        ("INDENT", 18),
        "arg, arg_context),",
        "POPINDENT",
        "POPINDENT",
        "patterns,",
        "arg_patterns)):",
        ("ENDING_LINENO", end_lineno),
        "POPINDENT",
        "rule.rule_base.num_bc_rules_matched += 1",
    )
    goal_fn_tail = (
        "POPINDENT",
        "POPINDENT",
        "finally:",
        ("INDENT", 2),
        "context.done()",
        "POPINDENT",
        "POPINDENT",
        "POPINDENT",
    )
    if not taking and not pred_plan_lines and not python_lines:
        plan_fn_name = "None"
        plan_lines = ()
    else:
        plan_fn_name = "%s_plans.%s" % (rb_name, rule_name)
        def_start = "def %s" % rule_name
        taking = [line.strip() for line in taking if line.strip()]
        if not taking:
            def_head = def_start + '(context):'
        else:
            if taking[0][0] != '(' or taking[-1][-1] != ')':
                from pyke.krb_compiler import scanner
                end = scanner.lexer.lexpos
                taking_start = scanner.lexer.lexdata.rfind('taking', 0, end)
                if taking_start < 0:
                    raise SyntaxError("'taking' clause: missing parenthesis",
                                      scanner.syntaxerror_params())
                taking_start += len('taking')
                while taking_start < len(scanner.lexdata) and \
                      scanner.lexdata[taking_start].isspace():
                    taking_start += 1
                lineno = scanner.lexer.lineno - \
                         scanner.lexer.lexdata.count('\n', taking_start,
                                                     scanner.lexer.lexpos)
                raise SyntaxError("'taking' clause: missing parenthesis",
                                  scanner.syntaxerror_params(taking_start,
                                                             lineno))
            taking[0] = def_start + "(context, " + taking[0][1:]
            taking[-1] += ':'
            if len(taking) == 1:
                def_head = taking[0]
            else:
                def_head = (taking[0],
                            ('INDENT', 4),
                            tuple(taking[1:]),
                            "POPINDENT",
                           )
        plan_lines = ("",
                      def_head,
                      ('INDENT', 2),
                      pred_plan_lines,
                      python_lines,
                      "POPINDENT",
                     )
    goal_decl_lines = (
        "",
        "bc_rule.bc_rule(%r, This_rule_base, %r," % (rule_name, goal_name),
        ("INDENT", 16),
        "%s, %s," % (rule_name, plan_fn_name),
    ) + list_format(pattern_args, "(", "),")
    return plan_lines, goal_fn_head, goal_fn_tail, goal_decl_lines

def add_start(l, start):
    '''
        >>> add_start(('a', 'b', 'c'), '^')
        (0, ['^a', 'b', 'c'])
        >>> add_start(('POPINDENT', ('INDENT', 2), ((('b',), 'c'),),), '^')
        (2, ['POPINDENT', ('INDENT', 2), ((('^b',), 'c'),)])
        >>> add_start((('POPINDENT', ('INDENT', 2)), ((('b',), 'c'),),), '^')
        (1, [('POPINDENT', ('INDENT', 2)), ((('^b',), 'c'),)])
        >>> add_start(('POPINDENT', ('INDENT', 2)), '^')
        (0, ['^', 'POPINDENT', ('INDENT', 2)])
    '''
    ans = list(l)
    for first, x in enumerate(ans):
        if x != 'POPINDENT' and \
           not (isinstance(x, (tuple, list)) and x[0] == 'INDENT'):
            if not isinstance(x, (tuple, list)):
                ans[first] = start + ans[first]
                return first, ans
            f, x2 = add_start(x, start)
            if len(x) == len(x2):
                ans[first] = tuple(x2)
                return first, ans
    first = 0
    ans.insert(first, start)
    return first, ans

def add_end(l, end):
    '''
        >>> add_end(('a', 'b', 'c'), '^')
        (2, ['a', 'b', 'c^'])
        >>> add_end(((((('b',), 'c'),),), 'POPINDENT', ('INDENT', 2)), '^')
        (0, [(((('b',), 'c^'),),), 'POPINDENT', ('INDENT', 2)])
        >>> add_end((((('b',), 'c'),), ('POPINDENT', ('INDENT', 2))), '^')
        (0, [((('b',), 'c^'),), ('POPINDENT', ('INDENT', 2))])
        >>> add_end(('POPINDENT', ('INDENT', 2)), '^')
        (2, ['POPINDENT', ('INDENT', 2), '^'])
    '''
    ans = list(l)
    for last in range(len(ans) - 1, -1, -1):
        x = ans[last]
        if x != 'POPINDENT' and \
           not (isinstance(x, (tuple, list)) and x[0] == 'INDENT'):
            if not isinstance(x, (tuple, list)):
                ans[last] += end
                return last, ans
            e, x2 = add_end(x, end)
            if len(x) == len(x2):
                ans[last] = tuple(x2)
                return last, ans
    last = len(ans)
    ans.insert(last, end)
    return last, ans

def add_brackets(l, start = '(', end = ')'):
    '''
        >>> add_brackets(('a', 'b', 'c'))
        ('(a', ('INDENT', 1), 'b', 'c)', 'POPINDENT')
        >>> add_brackets(('(a', ('INDENT', 1), 'b', 'c)', 'POPINDENT'))
        ('((a', ('INDENT', 1), ('INDENT', 1), 'b', 'c))', 'POPINDENT', 'POPINDENT')
    '''
    if not l: return start + end
    first, ans = add_start(l, start)
    last, ans = add_end(ans, end)
    if last > first:
        ans.insert(last + 1, "POPINDENT")
        ans.insert(first + 1, ("INDENT", 1))
    return tuple(ans)

def list_format(l, start, end, separator = ','):
    ans = [element + separator for element in l]
    if not ans: return (start + end,)
    ans[0] = start + ans[0]
    ans[-1] += end
    if len(ans) > 1:
        ans.insert(1, ("INDENT", 1))
        ans.append("POPINDENT")
    return tuple(ans)

def merge_pattern(pattern, pattern_list):
    # returns pat_num, new_pattern_list
    if pattern in pattern_list:
        return list(pattern_list).index(pattern), pattern_list
    return len(pattern_list), pattern_list + (pattern,)

def merge_patterns(patterns, pattern_list):
    # returns pat_nums, new_pattern_list
    pat_nums = []
    for pat in patterns:
        pat_num, pattern_list = merge_pattern(pat, pattern_list)
        pat_nums.append(pat_num)
    return tuple(pat_nums), pattern_list

def syntax_error(msg, lineno, pos):
    raise SyntaxError(msg, scanner.syntaxerror_params(pos, lineno))




# $Id: krb_traceback.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import linecache
import traceback
import os.path
import sys

def print_tb(traceback, limit=None, file=None):
    if file is None: file = sys.stderr
    for line in format_list(extract_tb(traceback, limit)): file.write(line)

def print_exception(type, value, traceback, limit=None, file=None):
    if file is None: file = sys.stderr
    if traceback:
        file.write('Traceback (most recent call last):\n')
        print_tb(traceback, limit, file)
    lines = format_exception_only(type, value)
    file.write(lines[0])
    for line in lines[1:]: file.write(' ' + line)

def print_exc(limit=None, file=None):
    type, value, traceback = sys.exc_info()
    print_exception(type, value, traceback, limit, file)

def format_exc(limit=None):
    type, value, traceback = sys.exc_info()
    return format_exception(type, value, traceback, limit)

def print_last(limit=None, file=None):
    print_exception(sys.last_type, sys.last_value, sys.last_traceback,
                    limit, file)

def print_stack(f=None, limit=None, file=None):
    if file is None: file = sys.stderr
    for line in format_list(extract_stack(f, limit)): file.write(line)

def extract_tb(tb, limit=None):
    ans = convert_tb(traceback.extract_tb(tb))
    if limit is not None and len(ans) > limit:
        return ans[len(ans) - limit:]
    return ans

def extract_stack(f=None, limit=None):
    ans = convert_tb(traceback.extract_stack(f))
    if limit is not None and len(ans) > limit:
        return ans[len(ans) - limit:]
    return ans

format_list = traceback.format_list

format_exception_only = traceback.format_exception_only

def format_exception(type, value, traceback, limit=None):
    ans = []
    if traceback:
        ans.append('Traceback (most recent call last):\n')
        ans.extend(format_list(extract_tb(traceback, limit)))
    lines = format_exception_only(type, value)
    ans.append(lines[0])
    for line in lines[1:]: ans.append(' ' + line)
    return ''.join(ans)

def format_tb(tb, limit=None):
    return format_list(extract_tb(tb, limit))

def format_stack(f=None, limit=None):
    return format_list(extract_stack(f, limit))

def convert_lineno(module, lineno):
    for (py_start, py_end), (krb_start, krb_end) in module.Krb_lineno_map:
        if py_start <= lineno <= py_end: return krb_start

def convert_tb(extracted_tb):
    '''
        extracted_tb is list of (filename, lineno, functionname, line)
    '''
    ans = []
    batch = []
    for info in extracted_tb:
        filename, lineno, functionname, line = info
        if filename.endswith('_fc.py') or filename.endswith('_bc.py'):
            pathname = filename[:-3]
            while True:
                module_name = pathname.replace(os.path.sep, '.')
                if module_name in sys.modules:
                    module = sys.modules[module_name]
                    if hasattr(module, 'Krb_filename'):
                        krb_lineno = convert_lineno(module, lineno)
                        if krb_lineno is not None:
                            if not ans: ans = batch
                            batch = []
                            krb_filename = \
                                os.path.normpath(
                                  os.path.join(os.path.dirname(module.__file__),
                                               module.Krb_filename))
                            linecache.checkcache(krb_filename)
                            line = linecache.getline(krb_filename, krb_lineno)
                            if line: line = line.strip()
                            else: line = None
                            ans.append((krb_filename, krb_lineno, functionname,
                                        line))
                            info = None
                        else:
                            ans.extend(batch)
                            ans.append(info)
                            batch = []
                            info = None
                    break
                sep_index = pathname.find(os.path.sep)
                if sep_index < 0: break
                pathname = pathname[sep_index + 1:]
        if info: batch.append(info)
    return ans + batch




# $Id: condensedPrint.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
import types
import re

def cprint(obj, maxlen = 80, maxdepth = 4, maxlines = 20):
    items = cprint2(obj, maxdepth)
    #sys.stderr.write("cprint items: %s\n" % str(items))
    return format(items, maxlen, maxlen, maxlines)[0]

def format_len(x):
    """
        >>> format_len('abc')
        3
        >>> format_len(('(', ('(', 'def', ')'), 'yz', ')'))
        11
    """
    if not isinstance(x, (list, tuple)): return len(x)
    if len(x) > 3: sep_len = 2 * (len(x) - 3)
    else: sep_len = 0
    return sum(map(format_len, x)) + sep_len

def format(x, lenleft, maxlen, maxlines, indent = 0):
    r"""
        >>> format('"hello mom this is a long str"', 7, 80, 9)
        ('"he..."', 0)
        >>> format(('(', 'a', 'b', 'c', ')'), 80, 80, 9)
        ('(a, b, c)', 0)
        >>> format(('(', 'a', 'b', 'c', ')'), 8, 80, 9)
        ('(a,\n  b,\n  c)', 2)
    """
    if not isinstance(x, (list, tuple)):
        if len(x) <= lenleft: return x, 0
        if isinstance(x, types.StringTypes) and x[-1] in "'\"":
            if lenleft >= 5: return x[:lenleft-4] + '...' + x[-1], 0
        else:
            if lenleft >= 4: return x[:lenleft-3] + '...', 0
        return '&', 0
    if len(x) == 0: return '', 0
    if format_len(x) <= lenleft:
        return x[0] + \
               ', '.join(format(y, lenleft, maxlen, maxlines)[0]
                           for y in x[1:-1]) + \
               x[-1], 0
    indent += 2
    ans = x[0]
    lines_taken = 0
    if len(x) > 2:
        first, taken = \
            format(x[1], lenleft - len(ans), maxlen, maxlines, indent + 2)
        ans += first
        lines_taken += taken
        for y in x[2:-1]:
            if lines_taken >= maxlines:
                ans += ', ...'
                break
            line, taken = \
                format(y, maxlen - indent, maxlen, maxlines - lines_taken,
                       indent)
            ans += ',\n' + indent * ' ' + line
            lines_taken += taken + 1
    return ans + x[-1], lines_taken

def cprint2(obj, maxdepth):
    if isinstance(obj, types.TupleType):
        return printSeq('(', ')', obj, maxdepth)
    if isinstance(obj, types.ListType):
        return printSeq('[', ']', obj, maxdepth)
    if isinstance(obj, types.DictType):
        return printDict(obj, maxdepth)
    if isinstance(obj, types.StringTypes):
        return printStr(obj)
    try:
        return str(obj)
    except StandardError, e:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        import traceback
        if isinstance(obj, types.InstanceType): obj_type = obj.__class__
        else: obj_type = type(obj)
        return "While trying to cprint a %s, got: %s" % \
                   (obj_type,
                    traceback.format_exception_only(exc_type, exc_value))

str_chk = re.compile('[a-zA-Z_][a-zA-Z0-9_]*$')

def printStr(str):
    """
        >>> printStr('hello_34_A')
        'hello_34_A'
        >>> printStr('hello 34_A')
        "'hello 34_A'"
    """
    if str_chk.match(str): return str
    return repr(str)

def printSeq(startChar, endChar, seq, maxdepth):
    """
        >>> printSeq('(', ')', (1, 2, 3), 4)
        ['(', '1', '2', '3', ')']
        >>> printSeq('(', ')', (), 4)
        ['(', ')']
    """
    if maxdepth < 1: return '&'
    maxdepth -= 1
    return [startChar] + [cprint2(x, maxdepth) for x in seq] + [endChar]

def item(key, value, maxdepth, separator):
    """
        >>> item('hello', 'bob', 3, '=')
        'hello=bob'
        >>> item(('hello', 'there'), 'bob', 3, '=')
        ['(', 'hello', 'there', ')=bob']
        >>> item('hello', ('extra', 'bob'), 3, '=')
        ['hello=(', 'extra', 'bob', ')']
        >>> item(('hello', 'there'), ('extra', 'bob'), 3, '=')
        ['(', 'hello', 'there', ')=(', 'extra', 'bob', ')']
    """
    keyans = cprint2(key, maxdepth)
    valans = cprint2(value, maxdepth)
    if isinstance(keyans, list):
        keyans[-1] += separator
        if isinstance(valans, list):
            keyans[-1] += valans[0]
            keyans.extend(valans[1:])
        else:
            keyans[-1] += valans
        return keyans
    if isinstance(valans, list):
        valans[0] = keyans + separator + valans[0]
        return valans
    return keyans + separator + valans

def printDict(dict, maxdepth,
                    startChar = '{', endChar = '}', separator = ': '):
    """
        >>> printDict({1:2, 3:4, 5:(6,7)}, 5)
        ['{', '1: 2', '3: 4', ['5: (', '6', '7', ')'], '}']
        >>> printDict({}, 5)
        ['{', '}']
    """
    if maxdepth < 1: return '&'
    maxdepth -= 1
    keys = dict.keys()
    keys.sort()
    return [startChar] + \
           [item(key, dict[key], maxdepth, separator) for key in keys] + \
           [endChar]




# $Id: fc_rule.py 38124415f0b5 2010-04-21 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

'''
    Forward chaining rules (fc_rule) are one of two types of rules in a
    rule_base (the other being backward chaining rules -- bc_rule).

    All forward chaining is done automatically as each rule_base is
    activated.  This is done in two steps:

        1.  All fc_rules are registered with the fact_lists referenced in
            their 'foreach' clause by calling fc_rule.register_rule() on
            each fc_rule (including the parent rule_base's fc_rules).

            This will cause the fact_list to invoke fc_rule.new_fact each time
            a new fact for that fact_list (by that name) is asserted (by the
            same or another fc_rule).

        2.  The fc_rule.run() function is called on each fc_rule (including
            the parent rule_base's fc_rules).

    The forward chaining rule is compiled into a python function which does
    the actual inferencing work for both the 'run' case and the 'new_fact'
    case, depending on the arguments passed to it.  Each fc_rule object
    remembers its associated compiled python function.

    The fc_rule object tracks the progress of the forward chaining for that
    rule.  If the rule has not been run yet, it ignores new_facts since it
    will see the new fact when it is later run.

'''

from pyke import contexts, fact_base

import itertools

class rule(object):
    ''' Common to both fc_rules and bc_rules. '''
    def __init__(self, name, rule_base, patterns):
        self.name = name
        self.rule_base = rule_base
        self.patterns = patterns

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.name)

    def pattern(self, pattern_index):
        return self.patterns[pattern_index]

class fc_rule(rule):
    def __init__(self, name, rule_base, rule_fn, foreach_facts, patterns):
        super(fc_rule, self).__init__(name, rule_base, patterns)
        rule_base.add_fc_rule(self)
        self.rule_fn = rule_fn
        self.foreach_facts = foreach_facts # (kb_name, fact_name, arg_pats,
                                           #  multi_match?)...
        self.ran = False

    def register_rule(self):
        for i, (kb_name, fact_name, arg_patterns, multi_match) \
         in enumerate(self.foreach_facts):
            self.rule_base.engine.get_kb(kb_name, fact_base.fact_base) \
                .add_fc_rule_ref(fact_name, self, i)

    def reset(self):
        self.ran = False

    def run(self):
        self.ran = True
        self.rule_fn(self)

    def new_fact(self, fact_args, n):
        if self.ran:
            arg_patterns = self.foreach_facts[n][2]
            if len(fact_args) == len(arg_patterns):
                context = contexts.simple_context()
                if all(itertools.imap(lambda pat, arg:
                                          pat.match_data(context, context, arg),
                                      arg_patterns,
                                      fact_args)):
                    self.rule_base.num_fc_rules_rerun += 1
                    if self.foreach_facts[n][3]:
                        self.rule_fn(self)
                    else:
                        self.rule_fn(self, context, n)
                context.done()

    def foreach_patterns(self, foreach_index):
        return self.foreach_facts[foreach_index][2]




# $Id: bc_rule.py 081917d30609 2010-03-05 mtnyogi $
# coding=utf-8
# 
# Copyright © 2007-2008 Bruce Frederiksen
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import functools
from pyke import fc_rule, immutable_dict

class bc_rule(fc_rule.rule):
    ''' This represents a single backward-chaining rule.  Most of its
        behavior is inherited.
    '''
    def __init__(self, name, rule_base, goal_name, bc_fn, plan_fn,
                 goal_arg_patterns, plan_vars, patterns):
        super(bc_rule, self).__init__(name, rule_base, patterns)
        self.goal_name = goal_name
        self.orig_bc_fn = bc_fn
        self.bc_fn = bc_fn
        self.plan_fn = plan_fn
        self.goal_arg_pats = goal_arg_patterns
        self.plan_vars = plan_vars
        rule_base.add_bc_rule(self)

    def goal_arg_patterns(self):
        return self.goal_arg_pats

    def make_plan(self, context, final):
        return functools.partial(self.plan_fn,
                   immutable_dict.immutable_dict(
                       (var_name, context.lookup_data(var_name, final=final))
                       for var_name in self.plan_vars))

    def trace(self):
        self.bc_fn = self.surrogate

    def surrogate(self, rule, arg_patterns, arg_context):
        print "%s.%s%s" % (rule.rule_base.root_name, rule.name,
                           tuple(arg.as_data(arg_context, True)
                                 for arg in arg_patterns))
        for prototype_plan in self.orig_bc_fn(rule, arg_patterns, arg_context):
            print "%s.%s succeeded with %s" % \
                  (rule.rule_base.root_name, rule.name,
                   tuple(arg.as_data(arg_context, True)
                         for arg in arg_patterns))
            yield prototype_plan
        print "%s.%s failed" % (rule.rule_base.root_name, rule.name)

    def untrace(self):
        self.bc_fn = self.orig_bc_fn

